"""
Basic test case for Gaphor tests.

Everything is about services so the TestCase can define it's required
services and start off.
"""

import logging
import unittest
from io import StringIO
from typing import Type, TypeVar

from gaphas.aspect import ConnectionSink
from gaphas.aspect import Connector as ConnectorAspect

# For DiagramItemConnector aspect:
import gaphor.diagram.diagramtools  # noqa
from gaphor import UML
from gaphor.application import Session
from gaphor.diagram.connectors import Connector
from gaphor.diagram.grouping import Group

T = TypeVar("T")

log = logging.getLogger("Gaphor")
log.setLevel(logging.WARNING)


class TestCase(unittest.TestCase):

    services = [
        "event_manager",
        "component_registry",
        "element_factory",
        "element_dispatcher",
        "modeling_language",
        "sanitizer",
    ]

    def setUp(self):
        self.session = Session(services=self.services)
        self.element_factory = self.session.get_service("element_factory")
        self.modeling_language = self.session.get_service("modeling_language")
        assert len(list(self.element_factory.select())) == 0, list(
            self.element_factory.select()
        )
        self.diagram = self.element_factory.create(UML.Diagram)
        assert len(list(self.element_factory.select())) == 1, list(
            self.element_factory.select()
        )

    def tearDown(self):
        self.element_factory.shutdown()
        self.session.shutdown()

    def get_service(self, name):
        return self.session.get_service(name)

    def create(self, item_cls: Type[T], subject_cls=None, subject=None) -> T:
        """
        Create an item with specified subject.
        """
        if subject_cls is not None:
            subject = self.element_factory.create(subject_cls)
        item = self.diagram.create(item_cls, subject=subject)
        self.diagram.canvas.update()
        return item

    def allow(self, line, handle, item, port=None):
        """
        Glue line's handle to an item.

        If port is not provided, then first port is used.
        """
        if port is None and len(item.ports()) > 0:
            port = item.ports()[0]

        adapter = Connector(item, line)
        return adapter.allow(handle, port)

    def connect(self, line, handle, item, port=None):
        """
        Connect line's handle to an item.

        If port is not provided, then first port is used.
        """
        canvas = line.canvas
        assert canvas is item.canvas
        if port is None and len(item.ports()) > 0:
            port = item.ports()[0]

        sink = ConnectionSink(item, port)
        connector = ConnectorAspect(line, handle)

        connector.connect(sink)

        cinfo = canvas.get_connection(handle)
        assert cinfo.connected is item
        assert cinfo.port is port

    def disconnect(self, line, handle):
        """
        Disconnect line's handle.
        """
        canvas = self.diagram.canvas
        # disconnection on adapter level is performed due to callback, so
        # no adapter look up here
        canvas.disconnect_item(line, handle)
        assert not canvas.get_connection(handle)

    def get_connected(self, handle):
        """
        Get item connected to line via handle.
        """
        cinfo = self.diagram.canvas.get_connection(handle)
        if cinfo:
            return cinfo.connected
        return None

    def get_connection(self, handle):
        """
        Get connection information.
        """
        return self.diagram.canvas.get_connection(handle)

    def can_group(self, parent, item):
        """
        Check if an item can be grouped by parent.
        """
        adapter = Group(parent, item)
        return adapter.can_contain()

    def group(self, parent, item):
        """
        Group item within a parent.
        """
        self.diagram.canvas.reparent(item, parent)
        adapter = Group(parent, item)
        adapter.group()

    def ungroup(self, parent, item):
        """
        Remove item from a parent.
        """
        adapter = Group(parent, item)
        adapter.ungroup()
        self.diagram.canvas.reparent(item, None)

    def kindof(self, cls):
        """
        Find UML metaclass instances using element factory.
        """
        return self.element_factory.lselect(cls)

    def save(self):
        """
        Save diagram into string.
        """
        from gaphor.storage import storage
        from gaphor.storage.xmlwriter import XMLWriter

        f = StringIO()
        storage.save(XMLWriter(f), factory=self.element_factory)
        data = f.getvalue()
        f.close()

        self.element_factory.flush()
        assert not list(self.element_factory.select())
        assert not list(self.element_factory.lselect())
        return data

    def load(self, data):
        """
        Load data from specified string. Update ``TestCase.diagram``
        attribute to hold new loaded diagram.
        """
        from gaphor.storage import storage

        f = StringIO(data)
        storage.load(
            f, factory=self.element_factory, modeling_language=self.modeling_language
        )
        f.close()

        self.diagram = self.element_factory.lselect(UML.Diagram)[0]



from gaphor.services.modelinglanguage import ModelingLanguageService


def test_loading_of_modeling_languages():
    modeling_language = ModelingLanguageService(event_manager=None, properties={})
    ids = [id for id, name in modeling_language.modeling_languages]
    assert "UML" in ids



from typing import Dict, List

import pytest

from gaphor.application import distribution
from gaphor.codegen.profile_coder import (
    create_referenced,
    filter_uml_classes,
    find_enumerations,
    generate,
    get_class_extensions,
    header,
    type_converter,
    write_attributes,
    write_properties,
)
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML import uml as UML
from gaphor.UML.classes import ClassItem
from gaphor.UML.classes.generalization import GeneralizationItem
from gaphor.UML.modelfactory import create_extension
from gaphor.UML.modelinglanguage import UMLModelingLanguage


class PseudoFile:
    def __init__(self):
        self.data = ""

    def write(self, data):
        self.data += data

    def close(self):
        pass


@pytest.fixture
def filename():
    return PseudoFile()


@pytest.fixture
def nodes(element_factory) -> Dict[int, ClassItem]:
    """Create tree using ClassItems and Stereotypes.

    All connections use UML.Generalizations, except for node 2 is connected to
    node 5 with an UML.Extension. Node 5 is the only Stereotype, the others are
    Classes.

    |- 0
    |  |- 1
    |  |- 2
    |  |  |- 3
    |  |  |- 4
    |- 5  |
    |  |- |- 6

    """
    diagram = element_factory.create(UML.Diagram)
    nodes = {
        n: diagram.create(ClassItem, subject=element_factory.create(UML.Class))
        for n in range(6)
    }
    nodes[6] = diagram.create(ClassItem, subject=element_factory.create(UML.Stereotype))
    gens = {
        n: diagram.create(
            GeneralizationItem, subject=element_factory.create(UML.Generalization)
        )
        for n in range(4)
    }
    class_connections = [0, 1, 0, 2, 2, 3, 2, 4]
    for n in range(4):
        connect(gens[n], gens[n].head, nodes[class_connections[n * 2]])
        connect(gens[n], gens[n].tail, nodes[class_connections[n * 2 + 1]])

    create_extension(nodes[2].subject, nodes[6].subject)
    create_extension(nodes[5].subject, nodes[6].subject)

    assert len(nodes) == 7
    return nodes


@pytest.fixture
def classes(nodes) -> List[UML.Class]:
    return [cls_item.subject for cls_item in nodes.values()]


def test_get_class_extension(classes):
    """Test getting meta class using UML.Extension."""
    meta_classes = get_class_extensions(classes[6])
    meta = [cls for cls in meta_classes]
    assert classes[2] in meta


def test_type_converter_with_type(element_factory):
    """Test type convert for boolean."""
    prop = element_factory.create(UML.Property)
    cls = element_factory.create(UML.Class)
    prop.type = cls
    prop.type.name = "boolean"
    assert type_converter(prop) == "boolean"


def test_type_converter_boolean(element_factory):
    """Test type convert for boolean."""
    prop = element_factory.create(UML.Property)
    prop.typeValue = "boolean"
    assert type_converter(prop) == "int"


def test_type_converter_int(element_factory):
    """Test type convert for an integer."""
    prop = element_factory.create(UML.Property)
    prop.typeValue = "integer"
    assert type_converter(prop) == "int"


def test_type_converter_str(element_factory):
    """Test type convert for a string."""
    prop = element_factory.create(UML.Property)
    prop.typeValue = "string"
    assert type_converter(prop) == "str"


def test_type_converter_other(element_factory):
    """Test type convert for other type."""
    prop = element_factory.create(UML.Property)
    prop.typeValue = "other_type"
    assert type_converter(prop) == "other_type"


def test_type_converter_none(element_factory):
    """Test type convert with no type."""
    prop = element_factory.create(UML.Property)
    with pytest.raises(ValueError):
        assert type_converter(prop)


def test_write_attributes_no_attribute(filename, element_factory):
    """Test writing pass when no attributes."""
    book = element_factory.create(UML.Class)

    write_attributes(book, filename)

    assert filename.data == "    pass\n\n"


def test_write_attributes_no_name(filename, element_factory):
    """Test writing pass when attribute has no name."""
    book = element_factory.create(UML.Class)
    book.ownedAttribute = element_factory.create(UML.Property)

    write_attributes(book, filename)

    assert filename.data == "    pass\n\n"


def test_write_attributes_for_attribute(filename, element_factory):
    """Test writing a normal attribute."""
    book = element_factory.create(UML.Class)
    book.ownedAttribute = element_factory.create(UML.Property)
    book.ownedAttribute[0].name = "title"
    book.ownedAttribute[0].typeValue = "str"

    write_attributes(book, filename)

    assert filename.data == "    title: attribute[str]\n"


def test_write_attributes_for_enumeration(filename, element_factory):
    """Test writing an attribute for an enumeration."""
    book = element_factory.create(UML.Class)
    book.ownedAttribute = element_factory.create(UML.Property)
    book.ownedAttribute[0].name = "duration"
    book.ownedAttribute[0].typeValue = "loanDurationKind"

    write_attributes(book, filename)

    assert filename.data == "    duration: enumeration\n"


def test_write_attributes_for_relation_one(filename, element_factory):
    """Test writing an attribute with upper value of one."""
    patron = element_factory.create(UML.Class)
    patron.ownedAttribute = element_factory.create(UML.Property)
    patron.ownedAttribute[0].name = "libraryCard"
    patron.ownedAttribute[0].typeValue = "Card"
    patron.ownedAttribute[0].upperValue = "1"

    write_attributes(patron, filename)

    assert filename.data == "    libraryCard: relation_one[Card]\n"


def test_write_attributes_for_relation_many(filename, element_factory):
    """Test writing an attribute with upper value of any."""
    library = element_factory.create(UML.Class)
    library.ownedAttribute = element_factory.create(UML.Property)
    library.ownedAttribute[0].name = "books"
    library.ownedAttribute[0].typeValue = "Book"
    library.ownedAttribute[0].upperValue = "*"

    write_attributes(library, filename)

    assert filename.data == "    books: relation_many[Book]\n"


def test_write_attributes_with_operations(filename, element_factory):
    """Test writing an attribute for an operation."""
    library = element_factory.create(UML.Class)
    library.ownedOperation = element_factory.create(UML.Operation)
    library.ownedOperation[0].name = "open"

    write_attributes(library, filename)

    assert filename.data == "    open: operation\n"


def test_filter_uml_classes(classes):
    """Test filtering of classes between UML and others."""
    classes[0].name = "~Class"
    classes[1].name = "Class"
    classes[2].name = "Behavior"
    classes[3].name = "Transportation"
    classes[4].name = "Car"
    classes[5].name = "Truck"
    classes[6].name = "Van"
    assert len(classes) == 7

    uml_classes = filter_uml_classes(classes, UMLModelingLanguage())

    assert len(uml_classes) == 2


def test_find_enumerations(classes):
    """Test filtering of classes between UML and others."""
    classes[0].name = "VehicleTypeKind"
    classes[1].name = "WheelKind"
    classes[2].name = "Behavior"
    classes[3].name = "Transportation"
    classes[4].name = "Car"
    classes[5].name = "Truck"
    classes[6].name = "Van"
    assert len(classes) == 7

    classes, enumerations = find_enumerations(classes)

    assert len(enumerations) == 2
    assert len(classes) == 5


def test_write_properties_no_attribute(filename, element_factory):
    """Test for no property written with no attributes."""
    book = element_factory.create(UML.Class)

    write_properties(book, filename)

    assert filename.data == ""


def test_write_properties_no_name(filename, element_factory):
    """Test writing a property when its attribute has no name."""
    book = element_factory.create(UML.Class)
    book.ownedAttribute = element_factory.create(UML.Property)

    write_properties(book, filename)

    assert filename.data == ""


def test_write_properties_for_attribute(filename, element_factory):
    """Test writing a property for a normal attribute."""
    book = element_factory.create(UML.Class)
    book.name = "Book"
    book.ownedAttribute = element_factory.create(UML.Property)
    book.ownedAttribute[0].name = "title"
    book.ownedAttribute[0].typeValue = "str"
    a = book.ownedAttribute[0]

    write_properties(book, filename)

    assert (
        filename.data
        == f'{book.name}.{a.name} = attribute("{a.name}", {a.typeValue})\n'
    )


def test_write_properties_for_enumeration(filename, element_factory):
    """Test writing property for an enumeration."""
    book = element_factory.create(UML.Class)
    book.name = "Book"
    book.ownedAttribute = element_factory.create(UML.Property)
    book.ownedAttribute[0].name = "duration"
    book.ownedAttribute[0].typeValue = "loanDurationKind"
    a = book.ownedAttribute[0]
    enum = element_factory.create(UML.Class)
    enum.name = "loanDurationKind"
    enum.ownedAttribute = element_factory.create(UML.Property)
    enum.ownedAttribute[0].name = "14"
    e = enum.ownedAttribute[0]
    enumerations = {enum.name: enum}

    write_properties(book, filename, enumerations)

    assert (
        filename.data
        == f'{book.name}.{a.name} = enumeration("kind", (\'{e.name}\',), "{e.name}")\n'
    )


def test_write_properties_for_association(filename, element_factory):
    """Test writing property for an association."""
    patron = element_factory.create(UML.Class)
    patron.ownedAttribute = element_factory.create(UML.Property)
    a = patron.ownedAttribute[0]
    a.name = "libraryCard"
    type_value = a.typeValue = "Card"
    a.upperValue = "1"
    a.aggregation = "composite"
    upper = f", upper={a.upperValue}"
    composite = ", composite=True"
    lower = opposite = ""

    write_properties(patron, filename)

    assert (
        filename.data
        == f'{patron.name}.{a.name} = association("{a.name}", {type_value}{lower}{upper}{composite}{opposite})\n'
    )


def test_create_referenced(classes):
    """Test list of referenced UML.Class objects."""
    referenced = create_referenced(classes)
    assert len(referenced) == 3

    nodes = [0, 2, 5]
    for node in nodes:
        assert classes[node] in referenced


def test_model_header(tmp_path):
    """Load a model with no relationships to test header."""
    path = distribution().locate_file("test-models/multiple-messages.gaphor")
    outfile = tmp_path / "profile.py"

    generate(path, outfile)

    assert header in outfile.read_text()


def test_model_with_extension(tmp_path):
    """Load a model with an extension relationship."""
    path = distribution().locate_file("test-models/codegen-extension.gaphor")
    outfile = tmp_path / "profile.py"

    generate(path, outfile)

    extension = """
from gaphor.UML import Class
class NewStereotype(Class):
    pass
"""
    assert extension in outfile.read_text()



"""
UI related events.
"""


class DiagramOpened:
    def __init__(self, diagram):
        self.diagram = diagram


class DiagramClosed:
    def __init__(self, diagram):
        self.diagram = diagram


class DiagramSelectionChanged:
    def __init__(self, diagram_view, focused_item, selected_items):
        self.diagram_view = diagram_view
        self.focused_item = focused_item
        self.selected_items = selected_items


class ModelingLanguageChanged:
    def __init__(self, modeling_language):
        self.modeling_language = modeling_language


class FileLoaded:
    def __init__(self, service, filename=None):
        self.service = service
        self.filename = filename


class FileSaved:
    def __init__(self, service, filename=None):
        self.service = service
        self.filename = filename



from gaphor.action import action


class ActionsMock:
    @action(name="simplest-action")
    def simplest_action(self):
        pass

    @action(name="tst.complete-action", label="a label", shortcut="shortcut-key")
    def complete_action(self):
        pass

    @action(name="param-action")
    def param_action(self, arg: str):
        pass


def test_simplest_action():
    action_data = ActionsMock.simplest_action.__action__

    assert action_data.name == "simplest-action"
    assert action_data.scope == "win"
    assert action_data.label is None


def test_complete_action():
    action_data = ActionsMock.complete_action.__action__

    assert action_data.name == "complete-action"
    assert action_data.scope == "tst"
    assert action_data.label == "a label"
    assert action_data.shortcut == "shortcut-key"


def test_param_action():
    action_data = ActionsMock.param_action.__action__

    assert action_data.name == "param-action"
    assert action_data.arg_type is str



import cairo
import pytest
from gaphas.geometry import Rectangle

from gaphor.core.modeling.diagram import FALLBACK_STYLE
from gaphor.diagram.shapes import (
    Box,
    DrawContext,
    IconBox,
    Text,
    TextAlign,
    VerticalAlign,
)


@pytest.fixture
def fixed_text_size(monkeypatch):
    size = (60, 15)

    def text_size(*args):
        return size

    monkeypatch.setattr("gaphor.diagram.text.Layout.size", text_size)
    return size


@pytest.fixture
def context():
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 0, 0)
    cr = cairo.Context(surface)
    return DrawContext(
        cairo=cr,
        style=FALLBACK_STYLE,
        selected=False,
        focused=False,
        hovered=False,
        dropzone=False,
    )


def test_box_size(context):
    box = Box()

    assert box.size(context=context) == (0, 0)


def test_draw_empty_box(context):
    box = Box(draw=None)

    box.draw(context=context, bounding_box=Rectangle())


def test_draw_box_with_custom_draw_function(context):
    called = False

    def draw(box, context, bounding_box):
        nonlocal called
        called = True

    box = Box(draw=draw)

    box.draw(context=context, bounding_box=Rectangle())

    assert called


def test_draw_icon_box(context):
    box_drawn = None

    def box_draw(box, context, bounding_box):
        nonlocal box_drawn
        box_drawn = bounding_box

    shape = IconBox(Box(draw=box_draw), Text(text="some text"))

    bounding_box = Rectangle(11, 12, 13, 14)
    shape.draw(context, bounding_box)

    assert box_drawn == bounding_box


def test_icon_box_child_placement_center_bottom(context):
    style = {"text-align": TextAlign.CENTER, "vertical-align": VerticalAlign.BOTTOM}

    text = Text(text="some text")
    shape = IconBox(Box(), text,)
    shape.size(context)

    w, h = shape.sizes[0]
    bounding_box = Rectangle(0, 0, 10, 20)

    x, y, _, _ = shape.child_pos(style, bounding_box)

    assert x == (bounding_box.width - w) / 2
    assert y == bounding_box.height


def test_icon_box_child_placement_right_middle(context):
    style = {"text-align": TextAlign.RIGHT, "vertical-align": VerticalAlign.MIDDLE}

    text = Text(text="some text")
    shape = IconBox(Box(), text,)
    shape.size(context)

    w, h = shape.sizes[0]
    bounding_box = Rectangle(0, 0, 10, 20)

    x, y, _, _ = shape.child_pos(style, bounding_box)

    assert x == bounding_box.width
    assert y == (bounding_box.height - h) / 2


def test_icon_box_child_placement_left_middle(context):
    style = {"text-align": TextAlign.LEFT, "vertical-align": VerticalAlign.MIDDLE}

    text = Text(text="some text")
    shape = IconBox(Box(), text,)
    shape.size(context)

    w, h = shape.sizes[0]
    bounding_box = Rectangle(0, 0, 10, 20)

    x, y, _, _ = shape.child_pos(style, bounding_box)

    assert x == -w
    assert y == (bounding_box.height - h) / 2


def test_icon_box_child_placement_center_top(context):
    style = {"text-align": TextAlign.CENTER, "vertical-align": VerticalAlign.TOP}

    text = Text(text="some text")
    shape = IconBox(Box(), text,)
    shape.size(context)

    w, h = shape.sizes[0]
    bounding_box = Rectangle(0, 0, 10, 20)

    x, y, _, _ = shape.child_pos(style, bounding_box)

    assert x == (bounding_box.width - w) / 2
    assert y == -h


def test_text_has_width(context, fixed_text_size):
    text = Text(lambda: "some text")

    w, _ = text.size(context)
    assert w == fixed_text_size[0]


def test_text_has_height(context, fixed_text_size):
    text = Text("some text")

    _, h = text.size(context)
    assert h == fixed_text_size[1]


def test_text_with_min_width(context):
    style = {"min-width": 100, "min-height": 0}
    text = Text("some text", style=style)

    w, _ = text.size(context)
    assert w == 100


def test_text_width_min_height(context):
    style = {"min-width": 0, "min-height": 40}
    text = Text("some text", style=style)

    _, h = text.size(context)
    assert h == 40



import abc
from typing import Tuple

from gaphor.abc import Service


class UIComponent(Service):
    """
    A user interface component.
    """

    ui_name: str  # "The UIComponent name, provided by the loader"

    title: str  # "Title of the component"

    size: Tuple[int, int]  # "Size used for floating the component"

    @abc.abstractmethod
    def open(self):
        """
        Create and display the UI components (windows).
        """

    @abc.abstractmethod
    def close(self):
        """
        Close the UI component. The component can decide to hide or destroy the UI
        components.
        """

    def shutdown(self):
        """
        Shut down this component. It's not supposed to be opened again.
        """
        self.close()



from typing import List, NamedTuple, Optional

from gaphor.diagram.copypaste import (
    ElementCopy,
    copy,
    copy_element,
    paste,
    paste_element,
)
from gaphor.UML import ExecutionSpecification, Message


class MessageCopy(NamedTuple):
    element_copy: ElementCopy
    send_event: Optional[ElementCopy]
    receive_event: Optional[ElementCopy]


@copy.register
def copy_message(element: Message):
    return MessageCopy(
        element_copy=copy_element(element),
        send_event=copy_element(element.sendEvent) if element.sendEvent else None,
        receive_event=copy_element(element.receiveEvent)
        if element.receiveEvent
        else None,
    )


@paste.register
def paste_message(copy_data: MessageCopy, diagram, lookup):
    if copy_data.send_event:
        paste_element(copy_data.send_event, diagram, lookup)
    if copy_data.receive_event:
        paste_element(copy_data.receive_event, diagram, lookup)
    return paste_element(copy_data.element_copy, diagram, lookup)


class ExecutionSpecificationCopy(NamedTuple):
    element_copy: ElementCopy
    occurrences: List[ElementCopy]


@copy.register
def copy_execution_specification(element: ExecutionSpecification):
    return ExecutionSpecificationCopy(
        element_copy=copy_element(element),
        occurrences=[
            copy_element(occurrence)
            for occurrence in element.executionOccurrenceSpecification
        ],
    )


@paste.register
def paste_execution_specification(
    copy_data: ExecutionSpecificationCopy, diagram, lookup
):
    for occurrence in copy_data.occurrences:
        paste_element(occurrence, diagram, lookup)
    return paste_element(copy_data.element_copy, diagram, lookup)



import pytest

from gaphor import UML
from gaphor.diagram.connectors import Connector
from gaphor.diagram.tests.fixtures import allow, connect, disconnect
from gaphor.SysML import sysml
from gaphor.SysML.blocks.block import BlockItem
from gaphor.SysML.blocks.connectors import BlockProperyProxyPortConnector
from gaphor.SysML.blocks.proxyport import ProxyPortItem
from gaphor.UML.components import ConnectorItem


@pytest.fixture
def block_item(diagram, element_factory):
    return diagram.create(BlockItem, subject=element_factory.create(sysml.Block))


@pytest.fixture
def proxy_port_item(diagram):
    return diagram.create(ProxyPortItem)


def connected_proxy_port_item(diagram, element_factory):
    proxy_port_item = diagram.create(ProxyPortItem)
    block_item = diagram.create(BlockItem, subject=element_factory.create(sysml.Block))

    connector = Connector(block_item, proxy_port_item)
    connector.connect(proxy_port_item.handles()[0], block_item.ports()[0])

    return proxy_port_item


@pytest.fixture
def head_proxy_port_item(diagram, element_factory):
    return connected_proxy_port_item(diagram, element_factory)


@pytest.fixture
def tail_proxy_port_item(diagram, element_factory):
    return connected_proxy_port_item(diagram, element_factory)


@pytest.fixture
def other_proxy_port_item(diagram, element_factory):
    return connected_proxy_port_item(diagram, element_factory)


@pytest.fixture
def connector_item(diagram):
    return diagram.create(ConnectorItem)


def test_connection_is_allowed(diagram, block_item, proxy_port_item):
    connector = Connector(block_item, proxy_port_item)

    assert isinstance(connector, BlockProperyProxyPortConnector)
    assert connector.allow(proxy_port_item.handles()[0], block_item.ports()[0])


def test_connect_proxy_port_to_block(diagram, block_item, proxy_port_item):
    connector = Connector(block_item, proxy_port_item)

    connected = connector.connect(proxy_port_item.handles()[0], block_item.ports()[0])

    assert connected
    assert proxy_port_item.subject
    assert proxy_port_item.subject.encapsulatedClassifier is block_item.subject
    assert proxy_port_item.subject in block_item.subject.ownedPort


def test_disconnect_proxy_port_to_block(diagram, block_item, proxy_port_item):
    connector = Connector(block_item, proxy_port_item)
    connector.connect(proxy_port_item.handles()[0], block_item.ports()[0])

    connector.disconnect(proxy_port_item.handles()[0])

    assert proxy_port_item.subject is None
    assert proxy_port_item.canvas


def test_allow_connector_to_proxy_port(
    diagram, connector_item: ConnectorItem, head_proxy_port_item: ProxyPortItem
):
    assert allow(connector_item, connector_item.handles()[0], head_proxy_port_item)


def test_connect_connector_on_one_end_to_proxy_port(
    diagram, connector_item: ConnectorItem, head_proxy_port_item: ProxyPortItem
):
    connect(connector_item, connector_item.handles()[0], head_proxy_port_item)

    assert connector_item.subject is None


def test_connect_connector_on_both_ends_to_proxy_port(
    connector_item: ConnectorItem,
    head_proxy_port_item: ProxyPortItem,
    tail_proxy_port_item: ProxyPortItem,
):
    connect(connector_item, connector_item.handles()[0], head_proxy_port_item)
    connect(connector_item, connector_item.handles()[1], tail_proxy_port_item)

    assert connector_item.subject
    assert head_proxy_port_item.subject in connector_item.subject.end[:].role
    assert tail_proxy_port_item.subject in connector_item.subject.end[:].role


def test_disconnect_connector_from_proxy_port(
    connector_item: ConnectorItem,
    head_proxy_port_item: ProxyPortItem,
    tail_proxy_port_item: ProxyPortItem,
    element_factory,
):
    connect(connector_item, connector_item.handles()[0], head_proxy_port_item)
    connect(connector_item, connector_item.handles()[1], tail_proxy_port_item)

    disconnect(connector_item, connector_item.handles()[0])

    assert not connector_item.subject
    assert element_factory.lselect(UML.Connector) == []
    assert element_factory.lselect(UML.ConnectorEnd) == []
    assert head_proxy_port_item.subject in element_factory.select(UML.Port)
    assert tail_proxy_port_item.subject in element_factory.select(UML.Port)



from gaphor import UML
from gaphor.diagram.tests.fixtures import copy_clear_and_paste
from gaphor.UML.states import StateItem


def test_state_with_behaviors(diagram, element_factory):
    state: UML.State = element_factory.create(UML.State)
    state.entry = element_factory.create(UML.Behavior)
    state.entry.name = "hi"
    state.exit = element_factory.create(UML.Behavior)
    state.exit.name = "bye"
    state.doActivity = element_factory.create(UML.Behavior)
    state.doActivity.name = "act"

    state_item = diagram.create(StateItem, subject=state)

    new_items = copy_clear_and_paste({state_item}, diagram, element_factory)
    new_state_item = new_items.pop()

    assert isinstance(new_state_item, StateItem)
    assert new_state_item.subject.entry.name == "hi"
    assert new_state_item.subject.exit.name == "bye"
    assert new_state_item.subject.doActivity.name == "act"



import pytest

from gaphor.application import Application
from gaphor.event import ActiveSessionChanged, SessionShutdown
from gaphor.ui import subscribe_to_lifecycle_events


class GtkApplicationStub:
    def __init__(self):
        self.has_quit = False

    def quit(self):
        self.has_quit = True


@pytest.fixture
def application():
    application = Application()
    yield application
    application.shutdown()


def two_sessions(application, gtk_app=GtkApplicationStub()):
    session1 = application.new_session(["event_manager"])
    subscribe_to_lifecycle_events(session1, application, gtk_app)

    session2 = application.new_session(["event_manager"])
    subscribe_to_lifecycle_events(session2, application, gtk_app)

    return session1, session2


def test_most_recently_created_session_is_active(application):
    session1, session2 = two_sessions(application)

    assert application.active_session is session2


def test_active_window_changed(application):
    session1, session2 = two_sessions(application)

    session1.get_service("event_manager").handle(ActiveSessionChanged(None))

    assert application.active_session is session1


def test_session_shutdown(application):
    session1, session2 = two_sessions(application)

    session2.get_service("event_manager").handle(SessionShutdown(None))

    assert len(application.sessions) == 1
    assert session1 in application.sessions


def test_all_sessions_shut_down(application):
    gtk_app = GtkApplicationStub()
    session1, session2 = two_sessions(application, gtk_app)

    session1.get_service("event_manager").handle(SessionShutdown(None))
    session2.get_service("event_manager").handle(SessionShutdown(None))

    assert len(application.sessions) == 0
    assert gtk_app.has_quit



from gaphor import UML
from gaphor.application import distribution
from gaphor.storage.parser import parse
from gaphor.storage.storage import load_elements
from gaphor.UML import diagramitems


def test_message_item_upgrade(element_factory, modeling_language):
    """
    """
    path = distribution().locate_file("test-models/multiple-messages.gaphor")

    elements = parse(path)
    load_elements(elements, element_factory, modeling_language)

    diagram = element_factory.lselect(UML.Diagram)[0]
    items = diagram.canvas.get_root_items()
    message_items = [i for i in items if isinstance(i, diagramitems.MessageItem)]
    subjects = [m.subject for m in message_items]
    messages = element_factory.lselect(UML.Message)
    presentations = [m.presentation for m in messages]

    assert len(messages) == 10
    assert all(subjects), subjects
    assert len(message_items) == 10
    assert all(presentations), presentations



"""
The main application window.
"""

import importlib.resources
import logging
from pathlib import Path
from typing import List, Tuple

from gi.repository import Gdk, Gio, GLib, Gtk

from gaphor.abc import ActionProvider, Service
from gaphor.core import action, event_handler, gettext
from gaphor.core.modeling import AttributeUpdated, Diagram, ModelFlushed, ModelReady
from gaphor.event import ActionEnabled, ActiveSessionChanged, SessionShutdownRequested
from gaphor.services.undomanager import UndoManagerStateChanged
from gaphor.ui.abc import UIComponent
from gaphor.ui.actiongroup import window_action_group
from gaphor.ui.diagrampage import DiagramPage
from gaphor.ui.event import (
    DiagramClosed,
    DiagramOpened,
    DiagramSelectionChanged,
    FileLoaded,
    FileSaved,
    ModelingLanguageChanged,
)
from gaphor.ui.layout import deserialize
from gaphor.ui.recentfiles import HOME, RecentFilesMenu

log = logging.getLogger(__name__)


def new_builder():
    builder = Gtk.Builder()
    builder.set_translation_domain("gaphor")
    with importlib.resources.path("gaphor.ui", "mainwindow.glade") as glade_file:
        builder.add_from_file(str(glade_file))
    return builder


def create_hamburger_model(export_menu, tools_menu):
    model = Gio.Menu.new()

    part = Gio.Menu.new()
    part.append(gettext("New Window"), "app.file-new")
    part.append(gettext("New from Template"), "app.file-new-template")
    model.append_section(None, part)

    part = Gio.Menu.new()
    part.append(gettext("Save"), "win.file-save")
    part.append(gettext("Save As..."), "win.file-save-as")
    part.append_submenu(gettext("Export"), export_menu)
    model.append_section(None, part)

    part = Gio.Menu.new()
    part.append_submenu(gettext("Tools"), tools_menu)
    model.append_section(None, part)

    part = Gio.Menu.new()
    part.append(gettext("Keyboard Shortcuts"), "app.shortcuts")
    part.append(gettext("About Gaphor"), "app.about")
    model.append_section(None, part)

    return model


def create_recent_files_model(recent_manager=None):
    model = Gio.Menu.new()
    model.append_section(
        gettext("Recently opened files"),
        RecentFilesMenu(recent_manager or Gtk.RecentManager.get_default()),
    )
    return model


def create_modeling_language_model(modeling_language):
    model = Gio.Menu.new()
    for id, name in modeling_language.modeling_languages:
        menu_item = Gio.MenuItem.new(name, "win.select-modeling-language")
        menu_item.set_attribute_value("target", GLib.Variant.new_string(id))
        model.append_item(menu_item)
    return model


class MainWindow(Service, ActionProvider):
    """
    The main window for the application.
    It contains a Namespace-based tree view and a menu and a statusbar.
    """

    size = property(lambda s: s.properties.get("ui.window-size", (760, 580)))

    def __init__(
        self,
        event_manager,
        component_registry,
        element_factory,
        properties,
        modeling_language,
        export_menu,
        tools_menu,
    ):
        self.event_manager = event_manager
        self.component_registry = component_registry
        self.element_factory = element_factory
        self.properties = properties
        self.modeling_language = modeling_language
        self.export_menu = export_menu
        self.tools_menu = tools_menu

        self.title = "Gaphor"
        self.window: Gtk.Window = None
        self.filename = None
        self.model_changed = False
        self.layout = None
        self.modeling_language_name = None

        self.init_styling()

    def init_styling(self):
        with importlib.resources.path("gaphor.ui", "layout.css") as css_file:
            style_provider = Gtk.CssProvider()
            style_provider.load_from_path(str(css_file))
            Gtk.StyleContext.add_provider_for_screen(
                Gdk.Screen.get_default(),
                style_provider,
                Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
            )

    def shutdown(self):
        if self.window:
            self.window.destroy()
            self.window = None

        em = self.event_manager
        em.unsubscribe(self._on_file_manager_state_changed)
        em.unsubscribe(self._on_undo_manager_state_changed)
        em.unsubscribe(self._new_model_content)
        em.unsubscribe(self._on_action_enabled)
        em.unsubscribe(self._on_modeling_language_selection_changed)

    def get_ui_component(self, name):
        return self.component_registry.get(UIComponent, name)

    def open(self, gtk_app=None):
        """Open the main window.
        """

        builder = new_builder()
        self.window = builder.get_object("main-window")
        self.window.set_application(gtk_app)

        select_modeling_language = builder.get_object("select-modeling-language")
        select_modeling_language.bind_model(
            create_modeling_language_model(self.modeling_language), None
        )
        self.modeling_language_name = builder.get_object("modeling-language-name")

        hamburger = builder.get_object("hamburger")
        hamburger.bind_model(
            create_hamburger_model(self.export_menu.menu, self.tools_menu.menu), None
        )

        recent_files = builder.get_object("recent-files")
        recent_files.bind_model(create_recent_files_model(), None)

        self.set_title()

        self.window.set_default_size(*self.size)

        def _factory(name):
            comp = self.get_ui_component(name)
            widget = comp.open()

            # Okay, this may be hackish. Action groups on component level are also added
            # to the main window. This ensures that we can call those items from the
            # (main) menus as well. Also this makes enabling/disabling work.
            for prefix in widget.list_action_prefixes():
                assert prefix not in ("app", "win")
                self.window.insert_action_group(prefix, widget.get_action_group(prefix))
            return widget

        with importlib.resources.open_text("gaphor.ui", "layout.xml") as f:
            self.layout = deserialize(self.window, f.read(), _factory, self.properties)

        action_group, accel_group = window_action_group(self.component_registry)
        self.window.insert_action_group("win", action_group)
        self.window.add_accel_group(accel_group)

        self._on_modeling_language_selection_changed()

        self.window.show_all()

        self.window.connect("notify::is-active", self._on_window_active)
        self.window.connect("delete-event", self._on_window_delete)

        # We want to store the window size, so it can be reloaded on startup
        self.window.set_resizable(True)
        self.window.connect("size-allocate", self._on_window_size_allocate)

        em = self.event_manager
        em.subscribe(self._on_file_manager_state_changed)
        em.subscribe(self._on_undo_manager_state_changed)
        em.subscribe(self._new_model_content)
        em.subscribe(self._on_action_enabled)
        em.subscribe(self._on_modeling_language_selection_changed)

    def open_welcome_page(self):
        """
        Create a new tab with a textual welcome page, a sort of 101 for
        Gaphor.
        """

    def set_title(self):
        """
        Sets the window title.
        """
        if not self.window:
            return

        if self.filename:
            p = Path(self.filename)
            title = p.name
            subtitle = str(p.parent).replace(HOME, "~")
        else:
            title = self.title
            subtitle = ""
        if self.model_changed:
            title += " [" + gettext("edited") + "]"
        self.window.set_title(title)
        self.window.get_titlebar().set_subtitle(subtitle)

    # Signal callbacks:

    @event_handler(ModelReady)
    def _new_model_content(self, event):
        """
        Open the toplevel element and load toplevel diagrams.
        """
        # TODO: Make handlers for ModelReady from within the GUI obj
        for diagram in self.element_factory.select(
            lambda e: e.isKindOf(Diagram)
            and not (e.namespace and e.namespace.namespace)
        ):
            self.event_manager.handle(DiagramOpened(diagram))

    @event_handler(FileLoaded, FileSaved)
    def _on_file_manager_state_changed(self, event):
        self.model_changed = False
        self.filename = event.filename
        self.set_title()

    @event_handler(UndoManagerStateChanged)
    def _on_undo_manager_state_changed(self, event):
        """
        """
        undo_manager = event.service
        if self.model_changed != undo_manager.can_undo():
            self.model_changed = undo_manager.can_undo()
            self.set_title()

    @event_handler(ActionEnabled)
    def _on_action_enabled(self, event):
        ag = self.window.get_action_group(event.scope)
        a = ag.lookup_action(event.name)
        a.set_enabled(event.enabled)

    @event_handler(ModelingLanguageChanged)
    def _on_modeling_language_selection_changed(self, event=None):
        if self.modeling_language_name:
            self.modeling_language_name.set_text(
                gettext("Profile: {}").format(
                    self.modeling_language.active_modeling_language_name
                )
            )

    def _on_window_active(self, window, prop):
        self.event_manager.handle(ActiveSessionChanged(self))

    def _on_window_delete(self, window, event):
        self.event_manager.handle(SessionShutdownRequested(self))
        return True

    def _on_window_size_allocate(self, window, allocation):
        """
        Store the window size in a property.
        """
        width, height = window.get_size()
        self.properties.set("ui.window-size", (width, height))


Gtk.AccelMap.add_filter("gaphor")


class Diagrams(UIComponent, ActionProvider):

    title = gettext("Diagrams")

    def __init__(self, event_manager, element_factory, properties, modeling_language):
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.properties = properties
        self.modeling_language = modeling_language
        self._notebook: Gtk.Notebook = None

    def open(self):
        """Open the diagrams component.

        Returns:
            The Gtk.Notebook.
        """

        self._notebook = Gtk.Notebook()
        self._notebook.props.scrollable = True
        self._notebook.show()
        self._notebook.connect("switch-page", self._on_switch_page)
        self.event_manager.subscribe(self._on_show_diagram)
        self.event_manager.subscribe(self._on_close_diagram)
        self.event_manager.subscribe(self._on_name_change)
        self.event_manager.subscribe(self._on_flush_model)
        return self._notebook

    def close(self):
        """Close the diagrams component."""

        self.event_manager.unsubscribe(self._on_flush_model)
        self.event_manager.unsubscribe(self._on_name_change)
        self.event_manager.unsubscribe(self._on_close_diagram)
        self.event_manager.unsubscribe(self._on_show_diagram)
        if self._notebook:
            self._notebook.destroy()
            self._notebook = None

    def get_current_diagram(self):
        """Returns the current page of the notebook.

        Returns (DiagramPage): The current diagram page.
        """

        page_num = self._notebook.get_current_page()
        child_widget = self._notebook.get_nth_page(page_num)
        if child_widget is not None:
            return child_widget.diagram_page.get_diagram()
        else:
            return None

    def get_current_view(self):
        """Returns the current view of the diagram page.

        Returns (GtkView): The current view.
        """
        if not self._notebook:
            return
        page_num = self._notebook.get_current_page()
        child_widget = self._notebook.get_nth_page(page_num)
        return child_widget and child_widget.diagram_page.get_view()

    def create_tab(self, title, widget):
        """Creates a new Notebook tab with a label and close button.

        Args:
            title (str): The title of the tab, the diagram name.
            widget (Gtk.Widget): The child widget of the tab.
        """

        page_num = self._notebook.append_page(
            child=widget, tab_label=self.tab_label(title, widget)
        )
        self._notebook.set_current_page(page_num)
        self._notebook.set_tab_reorderable(widget, True)

        view = widget.diagram_page.view
        self.event_manager.handle(
            DiagramSelectionChanged(view, view.focused_item, view.selected_items)
        )

    def tab_label(self, title, widget):
        tab_box = Gtk.Box.new(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        label = Gtk.Label.new(title)
        tab_box.pack_start(child=label, expand=True, fill=True, padding=0)

        close_image = Gtk.Image.new_from_icon_name(
            icon_name="window-close", size=Gtk.IconSize.BUTTON
        )
        button = Gtk.Button()
        button.set_relief(Gtk.ReliefStyle.NONE)

        # TODO: Call button.set_focus_on_click directly once PyGObject issue
        #  #371 is fixed
        Gtk.Widget.set_focus_on_click(button, False)

        button.add(close_image)
        button.connect(
            "clicked",
            lambda _button: self.event_manager.handle(
                DiagramClosed(widget.diagram_page.get_diagram())
            ),
        )
        tab_box.pack_start(child=button, expand=False, fill=False, padding=0)
        tab_box.show_all()
        return tab_box

    def get_widgets_on_pages(self):
        """Gets the widget on each open page Notebook page.

        The page is the page number in the Notebook (0 indexed) and the widget
        is the child widget on each page.

        Returns:
            List of tuples (page, widget) of the currently open Notebook pages.
        """

        widgets_on_pages: List[Tuple[int, Gtk.Widget]] = []
        if not self._notebook:
            return widgets_on_pages

        num_pages = self._notebook.get_n_pages()
        for page_num in range(num_pages):
            widget = self._notebook.get_nth_page(page_num)
            widgets_on_pages.append((page_num, widget))
        return widgets_on_pages

    def _on_switch_page(self, notebook, page, new_page_num):
        current_page_num = notebook.get_current_page()
        if current_page_num >= 0:
            self._clear_ui_settings(notebook.get_nth_page(current_page_num))
        self._add_ui_settings(page)
        view = page.diagram_page.view
        self.event_manager.handle(
            DiagramSelectionChanged(view, view.focused_item, view.selected_items)
        )

    def _add_ui_settings(self, page):
        window = page.get_toplevel()
        window.insert_action_group("diagram", page.action_group.actions)
        window.add_accel_group(page.action_group.shortcuts)

    def _clear_ui_settings(self, page):
        window = page.get_toplevel()
        window.insert_action_group("diagram", None)
        window.remove_accel_group(page.action_group.shortcuts)

    @action(name="close-current-tab", shortcut="<Primary>w")
    def close_current_tab(self):
        diagram = self.get_current_diagram()
        self.event_manager.handle(DiagramClosed(diagram))

    @event_handler(DiagramOpened)
    def _on_show_diagram(self, event):
        """Show a Diagram element in the Notebook.

        If a diagram is already open on a Notebook page, show that one,
        otherwise create a new Notebook page.

        Args:
            event: The service event that is calling the method.
        """

        diagram = event.diagram

        # Try to find an existing diagram page and give it focus
        for page, widget in self.get_widgets_on_pages():
            if widget.diagram_page.get_diagram() is diagram:
                self._notebook.set_current_page(page)
                self.get_current_view().grab_focus()
                return widget.diagram_page

        # No existing diagram page found, creating one
        page = DiagramPage(
            diagram,
            self.event_manager,
            self.element_factory,
            self.properties,
            self.modeling_language,
        )
        widget = page.construct()
        widget.diagram_page = page

        self.create_tab(diagram.name, widget)
        self.get_current_view().grab_focus()
        return page

    @event_handler(DiagramClosed)
    def _on_close_diagram(self, event):
        """Callback to close the tab and remove the notebook page.

        Args:
            button (Gtk.Button): The button the callback is from.
            widget (Gtk.Widget): The child widget of the tab.
        """
        diagram = event.diagram

        for page_num, widget in self.get_widgets_on_pages():
            if widget.diagram_page.get_diagram() is diagram:
                break
        else:
            log.warn(f"No tab found for diagram {diagram}")
            return

        if diagram is self.get_current_diagram():
            self._clear_ui_settings(widget)
        self._notebook.remove_page(page_num)
        widget.diagram_page.close()
        widget.destroy()

    @event_handler(ModelFlushed)
    def _on_flush_model(self, event):
        """
        Close all tabs.
        """
        while self._notebook.get_n_pages():
            self._notebook.remove_page(0)

    @event_handler(AttributeUpdated)
    def _on_name_change(self, event):
        if event.property is Diagram.name:
            for page in range(self._notebook.get_n_pages()):
                widget = self._notebook.get_nth_page(page)
                if event.element is widget.diagram_page.diagram:
                    self._notebook.set_tab_label(
                        widget, self.tab_label(event.new_value, widget)
                    )



"""
Object node item.
"""

import ast

from gaphas.state import observed, reversible_property

from gaphor import UML
from gaphor.diagram.presentation import ElementPresentation, Named
from gaphor.diagram.shapes import Box, EditableText, IconBox, Text, draw_border
from gaphor.diagram.support import represents
from gaphor.UML.modelfactory import stereotypes_str

DEFAULT_UPPER_BOUND = "*"


@represents(UML.ObjectNode)
class ObjectNodeItem(ElementPresentation, Named):
    """
    Representation of object node. Object node is ordered and has upper bound
    specification.

    Ordering information can be hidden by user.
    """

    def __init__(self, id=None, model=None):
        super().__init__(id, model)

        self._show_ordering = False

        self.shape = IconBox(
            Box(
                Text(text=lambda: stereotypes_str(self.subject),),
                EditableText(text=lambda: self.subject.name or ""),
                style={"min-width": 50, "min-height": 30, "padding": (5, 10, 5, 10)},
                draw=draw_border,
            ),
            Text(
                text=lambda: self.subject.upperBound not in (None, DEFAULT_UPPER_BOUND)
                and f"{{ upperBound = {self.subject.upperBound} }}",
            ),
            Text(
                text=lambda: self._show_ordering
                and self.subject.ordering
                and f"{{ ordering = {self.subject.ordering} }}",
            ),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")
        self.watch("subject[ObjectNode].upperBound")
        self.watch("subject[ObjectNode].ordering")

    @observed
    def _set_show_ordering(self, value):
        self._show_ordering = value
        self.request_update()

    show_ordering = reversible_property(lambda s: s._show_ordering, _set_show_ordering)

    def save(self, save_func):
        save_func("show-ordering", self._show_ordering)
        super().save(save_func)

    def load(self, name, value):
        if name == "show-ordering":
            self._show_ordering = ast.literal_eval(value)
        else:
            super().load(name, value)



from gaphor.diagram.grouping import AbstractGroup, Group
from gaphor.SysML.blocks.block import BlockItem
from gaphor.SysML.blocks.property import PropertyItem


@Group.register(BlockItem, PropertyItem)
class PropertyGroup(AbstractGroup):
    """
    Add Property to a Block.
    """

    def can_contain(self) -> bool:
        return (
            not self.item.subject.association
            or self.item.subject.owner is self.parent.subject
        )

    def group(self):
        if not self.item.subject.association:
            self.parent.subject.ownedAttribute = self.item.subject

    def ungroup(self):
        if not self.item.subject.association:
            del self.parent.subject.ownedAttribute[self.item.subject]



from __future__ import annotations

from typing import List, Optional

from gaphor.core import event_handler
from gaphor.core.modeling import Element
from gaphor.core.modeling.collection import collectionlist
from gaphor.core.modeling.event import AssociationUpdated
from gaphor.core.modeling.properties import (
    association,
    attribute,
    derived,
    derivedunion,
    enumeration,
    relation_many,
    relation_one,
)


def test_association_1_x():
    #
    # 1:-
    #
    class A(Element):
        one: relation_one[Optional[B]]

    class B(Element):
        two: relation_one[A]

    A.one = association("one", B, 0, 1, opposite="two")
    B.two = association("two", A, 0, 1)
    a = A()
    b = B()
    a.one = b
    assert a.one is b
    assert b.two is a

    a.one = None
    assert a.one is None
    assert b.two is None

    a.one = b
    assert a.one is b
    assert b.two is a

    del a.one
    assert a.one is None
    assert b.two is None


def test_association_n_x():
    #
    # n:-
    #
    class A(Element):
        one: relation_many[B]

    class B(Element):
        two: relation_one[A]

    A.one = association("one", B, 0, "*", opposite="two")
    B.two = association("two", A, 0, 1)

    a = A()
    b = B()
    a.one = b
    assert b in a.one
    assert b.two is a


def test_association_1_1():
    #
    # 1:1
    #
    class A(Element):
        one: relation_one[B]

    class B(Element):
        two: relation_one[A]

    class C(Element):
        pass

    A.one = association("one", B, 0, 1, opposite="two")
    B.two = association("two", A, 0, 1, opposite="one")

    a = A()
    b = B()
    a.one = b
    a.one = b

    assert a.one is b
    assert b.two is a

    a.one = B()
    assert a.one is not b
    assert b.two is None

    c = C()
    try:
        a.one = c
    except Exception:
        pass
    else:
        assert a.one is not c

    del a.one
    assert a.one is None
    assert b.two is None


def test_association_1_n():
    #
    # 1:n
    #
    class A(Element):
        one: relation_one[B]

    class B(Element):
        two: relation_many[A]

    class C(Element):
        pass

    A.one = association("one", B, lower=0, upper=1, opposite="two")
    B.two = association("two", A, lower=0, upper="*", opposite="one")

    a1 = A()
    a2 = A()
    b1 = B()
    b2 = B()

    b1.two = a1
    assert len(b1.two) == 1, "len(b1.two) == %d" % len(b1.two)
    assert a1 in b1.two
    assert a1.one is b1, f"{a1.one}/{b1}"
    b1.two = a1
    b1.two = a1
    assert len(b1.two) == 1, "len(b1.two) == %d" % len(b1.two)
    assert a1 in b1.two
    assert a1.one is b1, f"{a1.one}/{b1}"

    b1.two = a2
    assert a1 in b1.two
    assert a2 in b1.two
    assert a1.one is b1
    assert a2.one is b1

    try:
        del b1.two
    except Exception:
        pass  # ok
    else:
        assert b1.two != []

    assert a1 in b1.two
    assert a2 in b1.two
    assert a1.one is b1
    assert a2.one is b1

    b1.two.remove(a1)

    assert len(b1.two) == 1
    assert a1 not in b1.two
    assert a2 in b1.two
    assert a1.one is None
    assert a2.one is b1

    a2.one = b2

    assert len(b1.two) == 0
    assert len(b2.two) == 1
    assert a2 in b2.two
    assert a1.one is None
    assert a2.one is b2

    del b1.two[a1]


def test_association_n_n():
    #
    # n:n
    #
    class A(Element):
        one: relation_many[B]

    class B(Element):
        two: relation_many[A]

    class C(Element):
        pass

    A.one = association("one", B, 0, "*", opposite="two")
    B.two = association("two", A, 0, "*", opposite="one")

    a1 = A()
    a2 = A()
    b1 = B()
    b2 = B()

    a1.one = b1
    assert b1 in a1.one
    assert a1 in b1.two
    assert not a2.one
    assert not b2.two

    a1.one = b2
    assert b1 in a1.one
    assert b2 in a1.one
    assert a1 in b1.two
    assert a1 in b2.two
    assert not a2.one

    a2.one = b1
    assert len(a1.one) == 2
    assert len(a2.one) == 1
    assert len(b1.two) == 2
    assert len(b2.two) == 1
    assert b1 in a1.one
    assert b2 in a1.one
    assert a1 in b1.two
    assert a1 in b2.two
    assert b1 in a2.one
    assert a2 in b1.two

    del a1.one[b1]
    assert len(a1.one) == 1
    assert len(a2.one) == 1
    assert len(b1.two) == 1
    assert len(b2.two) == 1
    assert b1 not in a1.one
    assert b2 in a1.one
    assert a1 not in b1.two
    assert a1 in b2.two
    assert b1 in a2.one
    assert a2 in b1.two


def test_association_swap():
    class A(Element):
        one: relation_many[B]

    class B(Element):
        pass

    class C(Element):
        pass

    A.one = association("one", B, 0, "*")

    a = A()
    b1 = B()
    b2 = B()

    a.one = b1
    a.one = b2
    assert a.one.size() == 2
    assert a.one[0] is b1
    assert a.one[1] is b2

    events: List[object] = []

    @event_handler(AssociationUpdated)
    def handler(event, events=events):
        events.append(event)

    #        Application.register_handler(handler)
    #        try:
    a.one.swap(b1, b2)
    #            assert len(events) == 1
    #            assert events[0].property is A.one
    #            assert events[0].element is a
    #        finally:
    #            Application.unregister_handler(handler)

    assert a.one.size() == 2
    assert a.one[0] is b2
    assert a.one[1] is b1


def test_association_unlink_1():
    #
    # unlink
    #
    class A(Element):
        one: relation_many[B]

    class B(Element):
        pass

    class C(Element):
        pass

    A.one = association("one", B, 0, "*")

    a1 = A()
    a2 = A()
    b1 = B()
    b2 = B()

    a1.one = b1
    a1.one = b2
    assert b1 in a1.one
    assert b2 in a1.one

    a2.one = b1

    # remove b1 from all elements connected to b1
    # also the signal should be removed
    b1.unlink()

    assert b1 not in a1.one
    assert b2 in a1.one


def test_association_unlink_2():
    #
    # unlink
    #
    class A(Element):
        one: relation_many[B]

    class B(Element):
        two: relation_many[A]

    class C(Element):
        pass

    A.one = association("one", B, 0, "*", opposite="two")
    B.two = association("two", A, 0, "*")

    a1 = A()
    a2 = A()
    b1 = B()
    b2 = B()

    a1.one = b1
    a1.one = b2
    assert b1 in a1.one
    assert b2 in a1.one
    assert a1 in b1.two
    assert a1 in b2.two

    a2.one = b1

    # remove b1 from all elements connected to b1
    # also the signal should be removed
    b1.unlink()

    assert b1 not in a1.one
    assert b2 in a1.one
    assert a1 not in b1.two
    assert a1 in b2.two


def test_attributes():
    class A(Element):
        a: attribute[str]

    A.a = attribute("a", str, "default")

    a = A()
    assert a.a == "default", a.a
    a.a = "bar"
    assert a.a == "bar", a.a
    del a.a
    assert a.a == "default"
    try:
        a.a = 1
    except AttributeError:
        pass  # ok
    else:
        assert 0, "should not set integer"


def test_enumerations():
    class A(Element):
        a: enumeration

    A.a = enumeration("a", ("one", "two", "three"), "one")
    a = A()
    assert a.a == "one"
    a.a = "two"
    assert a.a == "two"
    a.a = "three"
    assert a.a == "three"
    try:
        a.a = "four"
    except AttributeError:
        assert a.a == "three"
    else:
        assert 0, "a.a could not be four"
    del a.a
    assert a.a == "one"


def test_derived():
    class A(Element):
        a: relation_many[A]

    A.a = derived(A, "a", str, 0, "*", lambda self: ["a", "b", "c"])

    a = A()

    assert isinstance(a.a, collectionlist)
    assert a.a == ["a", "b", "c"]


def test_derivedunion():
    class A(Element):
        a: relation_many[A]
        b: relation_one[A]
        u: relation_many[A]

    A.a = association("a", A)
    A.b = association("b", A, 0, 1)
    A.u = derivedunion(A, "u", object, 0, "*", A.a, A.b)

    a = A()
    assert len(a.a) == 0, f"a.a = {a.a}"
    assert len(a.u) == 0, f"a.u = {a.u}"
    a.a = b = A()
    a.a = c = A()
    assert len(a.a) == 2, f"a.a = {a.a}"
    assert b in a.a
    assert c in a.a
    assert len(a.u) == 2, f"a.u = {a.u}"
    assert b in a.u
    assert c in a.u

    a.b = d = A()
    assert len(a.a) == 2, f"a.a = {a.a}"
    assert b in a.a
    assert c in a.a
    assert d == a.b
    assert len(a.u) == 3, f"a.u = {a.u}"
    assert b in a.u
    assert c in a.u
    assert d in a.u


def test_derivedunion_notify_for_single_derived_property():
    class A(Element):
        pass

    class E(Element):
        notified = False

        a: relation_many[A]
        u: relation_many[A]

        def handle(self, event):
            if event.property is E.u:
                self.notified = True

    E.a = association("a", A)
    E.u = derivedunion(E, "u", A, 0, "*", E.a)

    e = E()
    e.a = A()

    assert e.notified is True


def test_derivedunion_notify_for_multiple_derived_properties():
    class A(Element):
        pass

    class E(Element):
        notified = False

        a: relation_many[A]
        aa: relation_many[A]
        u: relation_many[A]

        def handle(self, event):
            if event.property is E.u:
                self.notified = True

    E.a = association("a", A)
    E.aa = association("aa", A)
    E.u = derivedunion(E, "u", A, 0, "*", E.a, E.aa)

    e = E()
    e.a = A()

    assert e.notified is True


def test_derivedunion_listmixins():
    class A(Element):
        a: relation_many[A]
        b: relation_many[A]
        u: relation_many[A]
        name: attribute[str]

    A.a = association("a", A)
    A.b = association("b", A)
    A.u = derivedunion(A, "u", A, 0, "*", A.a, A.b)
    A.name = attribute("name", str, "default")

    a = A()
    a.a = A()
    a.a = A()
    a.b = A()
    a.a[0].name = "foo"
    a.a[1].name = "bar"
    a.b[0].name = "baz"

    assert list(a.a[:].name) == ["foo", "bar"]
    assert sorted(list(a.u[:].name)) == ["bar", "baz", "foo"]


def test_composite():
    class A(Element):
        is_unlinked = False
        name: str
        comp: relation_many[A]
        other: relation_many[A]

        def unlink(self):
            self.is_unlinked = True
            Element.unlink(self)

    A.comp = association("comp", A, composite=True, opposite="other")
    A.other = association("other", A, composite=False, opposite="comp")

    a = A()
    a.name = "a"
    b = A()
    b.name = "b"
    a.comp = b
    assert b in a.comp
    assert a in b.other

    a.unlink()
    assert a.is_unlinked
    assert b.is_unlinked



"""
Test connections to folded interface.
"""

from gaphor import UML
from gaphor.tests import TestCase
from gaphor.UML.classes.dependency import DependencyItem
from gaphor.UML.classes.implementation import ImplementationItem
from gaphor.UML.classes.interface import Folded, InterfaceItem
from gaphor.UML.classes.klass import ClassItem


class ImplementationTestCase(TestCase):
    def test_default_line_style(self):
        impl = self.create(ImplementationItem)

        assert impl.style["dash-style"]

    def test_folded_interface_connection(self):
        """Test connecting implementation to folded interface
        """
        iface = self.create(InterfaceItem, UML.Interface)
        iface.folded = Folded.PROVIDED
        impl = self.create(ImplementationItem)

        self.connect(impl, impl.head, iface, iface.ports()[0])

        assert not impl.style["dash-style"]

    def test_folded_interface_disconnection(self):
        """Test disconnection implementation from folded interface
        """
        iface = self.create(InterfaceItem, UML.Interface)
        iface.folded = Folded.PROVIDED
        impl = self.create(ImplementationItem)

        self.connect(impl, impl.head, iface, iface.ports()[0])
        self.disconnect(impl, impl.head)
        impl.request_update()

        assert impl.style["dash-style"]


class DependencyTestCase(TestCase):
    def test_default_line_style(self):
        dep = self.create(DependencyItem)

        assert dep.style["dash-style"]

    def test_folded_interface_connection(self):
        """Test connecting dependency to folded interface
        """
        clazz = self.create(ClassItem, UML.Class)
        iface = self.create(InterfaceItem, UML.Interface)
        iface.folded = Folded.PROVIDED
        dep = self.create(DependencyItem)

        self.connect(dep, dep.head, iface, iface.ports()[0])
        self.connect(dep, dep.tail, clazz, clazz.ports()[0])
        iface.request_update()
        iface.canvas.update_now()

        assert dep.subject
        assert not dep.style["dash-style"]
        assert iface.folded == Folded.REQUIRED

    def test_folded_interface_disconnection(self):
        """Test disconnection dependency from folded interface
        """
        iface = self.create(InterfaceItem, UML.Interface)
        iface.folded = Folded.PROVIDED
        dep = self.create(DependencyItem)

        self.connect(dep, dep.head, iface, iface.ports()[0])
        self.disconnect(dep, dep.head)
        dep.request_update()

        assert dep.style["dash-style"]
        assert iface.folded == Folded.PROVIDED

    def test_unfolded_interface_connection(self):
        """Test disconnection dependency from unfolded interface
        """
        iface = self.create(InterfaceItem, UML.Interface)
        dep = self.create(DependencyItem)

        self.connect(dep, dep.head, iface, iface.ports()[0])
        assert (7.0, 5.0) == dep.style["dash-style"]



from typing import NamedTuple, Optional

from gaphor.diagram.copypaste import (
    ElementCopy,
    copy,
    copy_element,
    paste,
    paste_element,
)
from gaphor.UML import State


class StateCopy(NamedTuple):
    element_copy: ElementCopy
    entry: Optional[ElementCopy]
    exit: Optional[ElementCopy]
    do_activity: Optional[ElementCopy]


@copy.register
def copy_state(element: State):
    return StateCopy(
        element_copy=copy_element(element),
        entry=copy_element(element.entry) if element.entry else None,
        exit=copy_element(element.exit) if element.exit else None,
        do_activity=copy_element(element.doActivity) if element.doActivity else None,
    )


@paste.register
def paste_state(copy_data: StateCopy, diagram, lookup):
    if copy_data.entry:
        paste_element(copy_data.entry, diagram, lookup)
    if copy_data.exit:
        paste_element(copy_data.exit, diagram, lookup)
    if copy_data.do_activity:
        paste_element(copy_data.do_activity, diagram, lookup)
    return paste_element(copy_data.element_copy, diagram, lookup)



from gaphor import UML
from gaphor.tests.testcase import TestCase
from gaphor.UML.actions.activitynodes import DecisionNodeItem, ForkNodeItem


class ActivityNodesTestCase(TestCase):
    def test_decision_node(self):
        """Test creation of decision node
        """
        self.create(DecisionNodeItem, UML.DecisionNode)

    def test_fork_node(self):
        """Test creation of fork node
        """
        self.create(ForkNodeItem, UML.ForkNode)

    def test_decision_node_persistence(self):
        """Test saving/loading of decision node
        """
        factory = self.element_factory
        item = self.create(DecisionNodeItem, UML.DecisionNode)

        data = self.save()
        self.load(data)

        item = self.diagram.canvas.select(DecisionNodeItem)[0]
        assert item.combined is None, item.combined

        merge_node = factory.create(UML.MergeNode)
        item.combined = merge_node
        data = self.save()
        self.load(data)

        item = self.diagram.canvas.select(DecisionNodeItem)[0]
        assert item.combined is not None, item.combined
        assert isinstance(item.combined, UML.MergeNode)

    def test_fork_node_persistence(self):
        """Test saving/loading of fork node
        """
        factory = self.element_factory
        item = self.create(ForkNodeItem, UML.ForkNode)

        data = self.save()
        self.load(data)

        item = self.diagram.canvas.select(ForkNodeItem)[0]
        assert item.combined is None, item.combined

        merge_node = factory.create(UML.JoinNode)
        item.combined = merge_node
        data = self.save()
        self.load(data)

        item = self.diagram.canvas.select(ForkNodeItem)[0]
        assert item.combined is not None, item.combined
        assert isinstance(item.combined, UML.JoinNode)



from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.shapes import Text, draw_arrow_head
from gaphor.diagram.support import represents
from gaphor.SysML import sysml
from gaphor.UML.modelfactory import stereotypes_str


class DirectedRelationshipPropertyPathItem(LinePresentation):

    relation_type = ""

    def __init__(self, id=None, model=None):
        super().__init__(id, model, style={"dash-style": (7.0, 5.0)})

        self.shape_middle = Text(
            text=lambda: stereotypes_str(self.subject, (self.relation_type,)),
            style={"min-width": 0, "min-height": 0},
        )
        self.draw_head = draw_arrow_head
        self.watch("subject.appliedStereotype.classifier.name")


@represents(sysml.Satisfy)
class SatisfyItem(DirectedRelationshipPropertyPathItem):

    relation_type = "satisfy"


@represents(sysml.DeriveReqt)
class DeriveReqtItem(DirectedRelationshipPropertyPathItem):

    relation_type = "deriveReqt"


@represents(sysml.Trace)
class TraceItem(DirectedRelationshipPropertyPathItem):

    relation_type = "trace"


@represents(sysml.Verify)
class VerifyItem(DirectedRelationshipPropertyPathItem):

    relation_type = "verify"


@represents(sysml.Refine)
class RefineItem(DirectedRelationshipPropertyPathItem):

    relation_type = "refine"



"""
Copy and paste data based on an element's `save()` and `load()` methods.

The `copy()` function will return all values serialized, either as string
values or reference id's.

The `paste()` function will resolve those values. Based on the elements
that are already in the model,

The copy() function returns only data that has to be part of the copy buffer.
the `paste()` function will load this data in a model.
"""

from __future__ import annotations

from functools import singledispatch
from typing import (
    Callable,
    Dict,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import gaphas

from gaphor.core.modeling import Diagram, Element, NamedElement, Presentation
from gaphor.core.modeling.collection import collection

T = TypeVar("T")


@singledispatch
def copy(obj: Element) -> T:
    """
    Create a copy of an element (or list of elements).
    The returned type should be distinct, so the `paste()`
    function can properly dispatch.
    """
    raise ValueError(f"No copier for {obj}")


@singledispatch
def paste(copy_data: T, diagram: Diagram, lookup: Callable[[str], Element]):
    """
    Paste previously copied data. Based on the data type created in the
    `copy()` function, try to duplicate the copied elements.
    """
    raise ValueError(f"No paster for {copy_data}")


def serialize(value):
    if isinstance(value, (Element, gaphas.Item)):
        return ("r", value.id)
    elif isinstance(value, collection):
        return ("c", [serialize(v) for v in value])
    else:
        if isinstance(value, bool):
            value = int(value)
        return ("v", str(value))


def deserialize(ser, lookup):
    vtype, value = ser
    if vtype == "r":
        e = lookup(value)
        if e:
            yield e
    elif vtype == "c":
        # TODO: should apply elements to model one at a time
        for v in value:
            yield from deserialize(v, lookup)
    elif vtype == "v":
        yield value


class ElementCopy(NamedTuple):
    cls: Type[Element]
    id: Union[str, bool]
    data: Dict[str, Tuple[str, str]]


def copy_element(element: Element) -> ElementCopy:
    buffer = {}

    def save_func(name, value):
        # do not copy Element.presentation, to avoid cyclic dependencies
        if name != "presentation":
            buffer[name] = serialize(value)

    element.save(save_func)
    return ElementCopy(cls=element.__class__, id=element.id, data=buffer)


copy.register(Element, copy_element)  # type: ignore[arg-type]


def paste_element(copy_data: ElementCopy, diagram, lookup):
    cls, id, data = copy_data
    element = diagram.model.create_as(cls, id)
    for name, ser in data.items():
        for value in deserialize(ser, lookup):
            element.load(name, value)
    element.postload()
    return element


paste.register(ElementCopy, paste_element)


class NamedElementCopy(NamedTuple):
    element_copy: ElementCopy
    with_namespace: bool


def copy_named_element(element: NamedElement) -> NamedElementCopy:
    return NamedElementCopy(
        element_copy=copy_element(element), with_namespace=bool(element.namespace)
    )


copy.register(NamedElement, copy_named_element)  # type: ignore[arg-type]


def paste_named_element(copy_data: NamedElementCopy, diagram, lookup):
    element = paste_element(copy_data.element_copy, diagram, lookup)
    if copy_data.with_namespace and not element.namespace:
        element.package = diagram.namespace
    return element


paste.register(NamedElementCopy, paste_named_element)


class PresentationCopy(NamedTuple):
    cls: Type[Element]
    data: Dict[str, Tuple[str, str]]
    parent: Optional[str]


@copy.register
def copy_presentation(item: Presentation) -> PresentationCopy:
    assert item.canvas
    buffer = {}

    def save_func(name, value):
        buffer[name] = serialize(value)

    item.save(save_func)
    parent = item.canvas.get_parent(item)
    return PresentationCopy(
        cls=item.__class__, data=buffer, parent=parent.id if parent else None
    )


@paste.register
def paste_presentation(copy_data: PresentationCopy, diagram, lookup):
    cls, data, parent = copy_data
    item = diagram.create(cls)
    if parent:
        p = lookup(parent)
        if p:
            diagram.canvas.reparent(item, p)
    for name, ser in data.items():
        for value in deserialize(ser, lookup):
            item.load(name, value)
    item.canvas.update_matrices([item])
    return item


class CopyData(NamedTuple):
    items: Dict[str, object]
    elements: Dict[str, object]


@copy.register
def _copy_all(items: set) -> CopyData:
    return CopyData(
        items={item.id: copy(item) for item in items},
        elements={
            item.subject.id: copy(item.subject)
            for item in items
            if isinstance(item, Presentation) and item.subject
        },
    )


@paste.register
def _paste_all(copy_data: CopyData, diagram, lookup) -> Set[Presentation]:
    new_items: Dict[str, Presentation] = {}
    new_elements: Dict[str, Optional[Element]] = {}

    # element_lookup prefers elements already in the model
    def element_lookup(ref: str):
        if ref in new_elements:
            return new_elements[ref]
        looked_up = lookup(ref)
        if looked_up:
            return looked_up
        if ref in copy_data.elements:
            new_elements[ref] = None
            new_elements[ref] = paste(copy_data.elements[ref], diagram, element_lookup)
            return new_elements[ref]

    # item_lookup copies items. Elements are looked up
    def item_lookup(ref: str):
        if ref in new_items:
            return new_items[ref]
        elif ref in copy_data.items:
            new_items[ref] = paste(copy_data.items[ref], diagram, item_lookup)
            return new_items[ref]
        looked_up = diagram.lookup(ref)
        if looked_up:
            return looked_up
        return element_lookup(ref)

    for old_id, data in copy_data.items.items():
        if old_id in new_items:
            continue
        new_items[old_id] = paste(data, diagram, item_lookup)

    for new_item in new_items.values():
        new_item.postload()

    return set(new_items.values())



from gaphor import UML
from gaphor.diagram.presentation import ElementPresentation, LinePresentation


class DummyVisualComponent:
    def size(self, ctx):
        return 0, 0

    def draw(self, ctx, bounding_box):
        pass


class StubElement(ElementPresentation):
    def __init__(self, id=None, model=None):
        super().__init__(id, model, shape=DummyVisualComponent())


class StubLine(LinePresentation):
    def __init__(self, id=None, model=None):
        super().__init__(id, model, shape_middle=DummyVisualComponent())


def test_creation(diagram):
    p = diagram.create(StubElement)

    assert p
    assert p.model
    assert p.subject is None


def test_element_sides(diagram):
    p = diagram.create(StubElement)

    port_top, port_right, port_bottom, port_left = p.ports()

    assert p.port_side(port_top) == "top"
    assert p.port_side(port_right) == "right"
    assert p.port_side(port_bottom) == "bottom"
    assert p.port_side(port_left) == "left"


def test_element_saving(element_factory, diagram):
    subject = element_factory.create(UML.Class)
    p = diagram.create(StubElement, subject=subject)

    properties = {}
    referenced = set()

    def save_func(name, value, reference=False):
        properties[name] = value
        if reference:
            referenced.add(name)

    p.save(save_func)

    assert len(properties) == 4
    assert properties["matrix"] == (1.0, 0.0, 0.0, 1.0, 0.0, 0.0)
    assert properties["width"] == 10.0
    assert properties["height"] == 10.0
    assert properties["subject"] is subject


def test_element_loading(element_factory, diagram):
    subject = element_factory.create(UML.Class)
    p = diagram.create(StubElement)

    p.load("matrix", "(2.0, 0.0, 0.0, 2.0, 0.0, 0.0)")
    p.load("width", "20")
    p.load("height", "25")
    p.load("subject", subject)

    assert tuple(p.matrix) == (2.0, 0.0, 0.0, 2.0, 0.0, 0.0)
    assert p.width == 20
    assert p.height == 25
    assert p.subject is subject


def test_line_saving(element_factory, diagram):
    subject = element_factory.create(UML.Dependency)
    p = diagram.create(StubLine, subject=subject)

    properties = {}
    referenced = set()

    def save_func(name, value, reference=False):
        properties[name] = value
        if reference:
            referenced.add(name)

    p.save(save_func)

    assert properties["matrix"] == (1.0, 0.0, 0.0, 1.0, 0.0, 0.0)
    assert properties["orthogonal"] is False
    assert properties["horizontal"] is False
    assert properties["points"] == [(0.0, 0.0), (10.0, 10.0)]
    assert properties["subject"] is subject
    assert "head-connection" not in properties
    assert "tail-connection" not in properties


def test_line_saving_without_subject(diagram):
    p = diagram.create(StubLine)

    properties = {}
    referenced = set()

    def save_func(name, value, reference=False):
        properties[name] = value
        if reference:
            referenced.add(name)

    p.save(save_func)

    assert "subject" not in properties


def test_line_loading(element_factory, diagram):

    with element_factory.block_events():
        subject = element_factory.create(UML.Dependency)
        p = diagram.create(StubLine)

        p.load("matrix", "(2.0, 0.0, 0.0, 2.0, 0.0, 0.0)")
        p.load("orthogonal", "0")
        p.load("horizontal", "1")
        p.load("points", "[(1.0, 2.0), (3.0, 4.0)]")
        p.load("subject", subject)

    assert tuple(p.matrix) == (2.0, 0.0, 0.0, 2.0, 0.0, 0.0)
    assert not p.orthogonal
    assert p.horizontal
    assert tuple(p.handles()[0].pos) == (1.0, 2.0)
    assert tuple(p.handles()[1].pos) == (3.0, 4.0)
    assert p.subject is subject



"""Test Stereotype Property Page."""

import pytest
from gi.repository import Gtk

from gaphor import UML
from gaphor.UML.classes.klass import ClassItem
from gaphor.UML.profiles.stereotypepropertypages import StereotypePage


@pytest.fixture
def class_(diagram, element_factory):
    class_ = diagram.create(ClassItem, subject=element_factory.create(UML.Class))
    class_.name = "Class"
    yield class_
    del class_


def test_stereotype_page_with_no_stereotype(diagram, class_):
    """Test the Stereotype Property Page not created for a Class."""

    editor = StereotypePage(class_)
    page = editor.construct()

    assert page is None


def test_stereotype_page_with_stereotype(element_factory, diagram, class_):
    """Test creation of a Stereotype Property Page."""

    # Create a stereotype applicable to Class types:
    metaclass = element_factory.create(UML.Class)
    metaclass.name = "Class"
    stereotype = element_factory.create(UML.Stereotype)
    stereotype.name = "NewStereotype"
    UML.model.create_extension(metaclass, stereotype)
    attr = element_factory.create(UML.Property)
    attr.name = "Property"

    editor = StereotypePage(class_)
    page = editor.construct()

    box = page.get_children()[0]
    frame = box.get_children()[1]
    stereotype_view = frame.get_children()[0]

    assert isinstance(stereotype_view, Gtk.TreeView)
    assert len(stereotype_view.get_model()) == 1
    assert page is not None



"""
Test classifier stereotypes attributes using component items.
"""

from gaphor import UML
from gaphor.tests import TestCase
from gaphor.UML.components.component import ComponentItem


def compartments(item):
    return item.shape.children[1:]


class StereotypesAttributesTestCase(TestCase):
    def setUp(self):
        """
        Create two stereotypes and extend component UML metaclass using
        them.
        """
        super().setUp()
        factory = self.element_factory
        cls = factory.create(UML.Class)
        cls.name = "Component"
        st1 = self.st1 = factory.create(UML.Stereotype)
        st1.name = "st1"
        st2 = self.st2 = factory.create(UML.Stereotype)
        st2.name = "st2"

        attr = factory.create(UML.Property)
        attr.name = "st1_attr_1"
        st1.ownedAttribute = attr
        attr = factory.create(UML.Property)
        attr.name = "st1_attr_2"
        st1.ownedAttribute = attr

        attr = factory.create(UML.Property)
        attr.name = "st2_attr_1"
        st2.ownedAttribute = attr

        self.ext1 = UML.model.create_extension(cls, st1)
        self.ext2 = UML.model.create_extension(cls, st2)

    def tearDown(self):
        del self.st1
        del self.st2

    def test_adding_slot(self):
        """Test if stereotype attribute information is added when slot is added
        """
        c = self.create(ComponentItem, UML.Component)

        c.show_stereotypes = True
        instance_spec = UML.model.apply_stereotype(c.subject, self.st1)

        # test precondition
        assert not compartments(c)

        slot = UML.model.add_slot(instance_spec, self.st1.ownedAttribute[0])
        slot.value = "foo"

        assert len(compartments(c)) == 1

    def test_removing_last_slot(self):
        """Test removing last slot
        """
        c = self.create(ComponentItem, UML.Component)

        c.show_stereotypes = True
        instance_spec = UML.model.apply_stereotype(c.subject, self.st1)

        slot = UML.model.add_slot(instance_spec, self.st1.ownedAttribute[0])
        slot.value = "foo"

        # test precondition
        assert compartments(c)

        del instance_spec.slot[slot]
        assert not compartments(c)

    def test_deleting_extension(self):
        """Test if stereotype is removed when extension is deleteded
        """
        c = self.create(ComponentItem, UML.Component)

        c.show_stereotypes = True

        instance_spec = UML.model.apply_stereotype(c.subject, self.st1)
        slot = UML.model.add_slot(instance_spec, self.st1.ownedAttribute[0])
        slot.value = "foo"

        # test precondition
        assert len(compartments(c)) == 1
        assert len(c.subject.appliedStereotype) == 1

        self.ext1.unlink()
        assert len(c.subject.appliedStereotype) == 0
        assert len(compartments(c)) == 0

    def test_deleting_stereotype(self):
        """Test if stereotype is removed when stereotype is deleteded
        """
        c = self.create(ComponentItem, UML.Component)

        c.show_stereotypes = True

        st1 = self.st1
        instance_spec = UML.model.apply_stereotype(c.subject, st1)
        slot = UML.model.add_slot(instance_spec, self.st1.ownedAttribute[0])
        slot.value = "foo"

        # test precondition
        assert len(compartments(c)) == 1
        assert len(c.subject.appliedStereotype) == 1

        st1.unlink()
        assert len(c.subject.appliedStereotype) == 0
        assert len(compartments(c)) == 0

    def test_removing_stereotype_attribute(self):
        """Test if stereotype instance specification is destroyed when stereotype attribute is removed
        """
        c = self.create(ComponentItem, UML.Component)

        c.show_stereotypes = True

        # test precondition
        assert len(compartments(c)) == 0
        obj = UML.model.apply_stereotype(c.subject, self.st1)
        # test precondition
        assert len(compartments(c)) == 0

        assert len(self.kindof(UML.Slot)) == 0

        attr = self.st1.ownedAttribute[0]
        slot = UML.model.add_slot(obj, attr)
        slot.value = "foo"
        assert len(obj.slot) == 1
        assert len(self.kindof(UML.Slot)) == 1
        assert slot.definingFeature

        assert compartments(c)

        attr.unlink()
        assert len(obj.slot) == 0
        assert 0 == len(self.kindof(UML.Slot))
        assert not compartments(c)

    def test_stereotype_attributes_status_saving(self):
        """Test stereotype attributes status saving
        """
        c = self.create(ComponentItem, UML.Component)

        c.show_stereotypes = True
        UML.model.apply_stereotype(c.subject, self.st1)
        obj = UML.model.apply_stereotype(c.subject, self.st2)

        # change attribute of 2nd stereotype
        attr = self.st2.ownedAttribute[0]
        slot = UML.model.add_slot(obj, attr)
        slot.value = "st2 test21"

        data = self.save()
        self.load(data)

        item = self.diagram.canvas.select(ComponentItem)[0]
        assert item.show_stereotypes
        assert len(compartments(c)) == 1

    def test_saving_stereotype_attributes(self):
        """Test stereotype attributes saving
        """
        c = self.create(ComponentItem, UML.Component)

        c.show_stereotypes = True

        UML.model.apply_stereotype(c.subject, self.st1)
        UML.model.apply_stereotype(c.subject, self.st2)

        assert len(self.st1.ownedAttribute) == 3
        attr1, attr2, attr3 = self.st1.ownedAttribute
        assert attr1.name == "st1_attr_1", attr1.name
        assert attr2.name == "st1_attr_2", attr2.name
        assert attr3.name == "baseClass", attr3.name

        obj = c.subject.appliedStereotype[0]
        slot = UML.model.add_slot(obj, attr1)
        slot.value = "st1 test1"
        slot = UML.model.add_slot(obj, attr2)
        slot.value = "st1 test2"

        data = self.save()
        self.load(data)

        item = self.diagram.canvas.select(ComponentItem)[0]
        el = item.subject
        assert len(el.appliedStereotype) == 2

        # check if stereotypes are properly applied
        names = sorted(obj.classifier[0].name for obj in el.appliedStereotype)
        assert ["st1", "st2"] == names

        # two attributes were changed for stereotype st1, so 2 slots
        obj = el.appliedStereotype[0]
        assert len(obj.slot) == 2
        assert "st1_attr_1" == obj.slot[0].definingFeature.name
        assert "st1 test1" == obj.slot[0].value
        assert "st1_attr_2" == obj.slot[1].definingFeature.name
        assert "st1 test2" == obj.slot[1].value

        # no stereotype st2 attribute changes, no slots
        obj = el.appliedStereotype[1]
        assert len(obj.slot) == 0



from gaphor.UML.classes import DependencyItem
from gaphor.UML.components import NodeItem

PARENT_X = 189
PARENT_Y = 207
CHILD_ONE_X = 32
CHILD_ONE_Y = 54
CHILD_TWO_X = 44
CHILD_TWO_Y = 208


def handle_pos(item, handle_index):
    return tuple(map(float, item.handles()[handle_index].pos))


def test_load_grouped_connected_items(element_factory, loader):
    loader(NODE_EXAMPLE_XML)

    diagram = element_factory.lselect()[0]
    canvas = diagram.canvas
    node_item, dep_item = canvas.get_root_items()
    child_one, child_two = canvas.get_children(node_item)

    assert isinstance(node_item, NodeItem)
    assert isinstance(dep_item, DependencyItem)
    assert isinstance(child_one, NodeItem)
    assert isinstance(child_two, NodeItem)

    assert canvas.get_parent(child_one) is node_item

    assert tuple(canvas.get_matrix_i2c(child_one)) == (
        1.0,
        0.0,
        0.0,
        1.0,
        PARENT_X + CHILD_ONE_X,
        PARENT_Y + CHILD_ONE_Y,
    )
    assert tuple(canvas.get_matrix_i2c(child_two)) == (
        1.0,
        0.0,
        0.0,
        1.0,
        PARENT_X + CHILD_TWO_X,
        PARENT_Y + CHILD_TWO_Y,
    )


NODE_EXAMPLE_XML = f"""\
<?xml version="1.0" encoding="utf-8"?>
<gaphor xmlns="http://gaphor.sourceforge.net/model" version="3.0" gaphor-version="1.2.0rc2">
  <Diagram id="3ea414eb-5eb5-11ea-9ccf-45e3771927d8">
    <canvas>
      <item id="41f681ab-5eb5-11ea-9ccf-45e3771927d8" type="NodeItem">
        <matrix>
          <val>(1.0, 0.0, 0.0, 1.0, {PARENT_X}, {PARENT_Y})</val>
        </matrix>
        <width>
          <val>388.5</val>
        </width>
        <height>
          <val>286.5</val>
        </height>
        <item id="4541c555-5eb5-11ea-9ccf-45e3771927d8" type="NodeItem">
          <matrix>
            <val>(1.0, 0.0, 0.0, 1.0, {CHILD_ONE_X}, {CHILD_ONE_Y})</val>
          </matrix>
          <width>
            <val>100.0</val>
          </width>
          <height>
            <val>50.0</val>
          </height>
        </item>
        <item id="4f927913-5eb5-11ea-9ccf-45e3771927d8" type="NodeItem">
          <matrix>
            <val>(1.0, 0.0, 0.0, 1.0, {CHILD_TWO_X}, {CHILD_TWO_Y})</val>
          </matrix>
          <width>
            <val>100.0</val>
          </width>
          <height>
            <val>50.0</val>
          </height>
        </item>
      </item>
      <item id="5b4d81cb-5eb5-11ea-9ccf-45e3771927d8" type="DependencyItem">
        <matrix>
          <val>(1.0, 0.0, 0.0, 1.0, 274.0, 312.0)</val>
        </matrix>
        <points>
          <val>[(0.0, 0.0), (6.0, 104.0)]</val>
        </points>
        <head-connection>
          <ref refid="4541c555-5eb5-11ea-9ccf-45e3771927d8"/>
        </head-connection>
        <tail-connection>
          <ref refid="4f927913-5eb5-11ea-9ccf-45e3771927d8"/>
        </tail-connection>
      </item>
    </canvas>
  </Diagram>
</gaphor>
"""



from gaphor.diagram.grouping import AbstractGroup, Group
from gaphor.UML.interactions.interaction import InteractionItem
from gaphor.UML.interactions.lifeline import LifelineItem


@Group.register(InteractionItem, LifelineItem)
class InteractionLifelineGroup(AbstractGroup):
    """
    Add lifeline to interaction.
    """

    def group(self):
        assert self.parent.canvas
        self.parent.subject.lifeline = self.item.subject
        self.parent.canvas.reparent(self.item, self.parent)

    def ungroup(self):
        del self.parent.subject.lifeline[self.item.subject]



import inspect
import logging
from typing import Dict, Type, TypeVar

import importlib_metadata

T = TypeVar("T")

logger = logging.getLogger(__name__)


def initialize(scope, services=None, **known_services: T) -> Dict[str, T]:
    return init_entrypoints(load_entrypoints(scope, services), **known_services)


def load_entrypoints(scope, services=None) -> Dict[str, Type[T]]:
    """
    Load services from resources.
    """
    uninitialized_services = {}
    for ep in importlib_metadata.entry_points()[scope]:
        cls = ep.load()
        if not services or ep.name in services:
            logger.debug(f'found service entry point "{ep.name}"')
            uninitialized_services[ep.name] = cls
    return uninitialized_services


def init_entrypoints(
    uninitialized_services: Dict[str, Type[T]], **known_services: T
) -> Dict[str, T]:
    """
    Instantiate service definitions, taking into account dependencies
    defined in the constructor.

    Given a dictionary `{name: service-class}`,
    return a map `{name: service-instance}`.
    """
    ready: Dict[str, T] = dict(known_services)

    def pop(name):
        try:
            return uninitialized_services.pop(name)
        except KeyError:
            return None

    def init(name, cls):
        kwargs = {}
        for dep in inspect.signature(cls).parameters:
            if dep not in ready:
                depcls = pop(dep)
                if depcls:
                    kwargs[dep] = init(dep, depcls)
                else:
                    logger.debug(
                        f"Entrypont {name} parameter {dep} does not reference a resolved dependency"
                    )
            else:
                kwargs[dep] = ready[dep]
        srv = cls(**kwargs)
        ready[name] = srv
        return srv

    while uninitialized_services:
        name = next(iter(uninitialized_services.keys()))
        cls = pop(name)
        init(name, cls)

    return ready



"""
Stereotype property page.
"""

from gi.repository import Gtk

from gaphor import UML
from gaphor.core import transactional
from gaphor.diagram.propertypages import PropertyPageBase, PropertyPages, new_builder


@PropertyPages.register(UML.Element)
class StereotypePage(PropertyPageBase):

    order = 40

    def __init__(self, item):
        self.item = item

    def construct(self):

        subject = self.item.subject
        if not subject:
            return None

        stereotypes = UML.model.get_stereotypes(subject)
        if not stereotypes:
            return None

        builder = new_builder("stereotypes-editor")

        show_stereotypes = builder.get_object("show-stereotypes")

        if hasattr(self.item, "show_stereotypes"):
            show_stereotypes.set_active(self.item.show_stereotypes)
        else:
            show_stereotypes.destroy()

        model, toggle_stereotype_handler, set_slot_value_handler = stereotype_model(
            subject
        )

        stereotype_list = builder.get_object("stereotype-list")
        stereotype_list.set_model(model)

        builder.connect_signals(
            {
                "show-stereotypes-changed": (self._on_show_stereotypes_change,),
                "toggle-stereotype": toggle_stereotype_handler,
                "set-slot-value": set_slot_value_handler,
            }
        )

        return builder.get_object("stereotypes-editor")

    @transactional
    def _on_show_stereotypes_change(self, button, gparam):
        self.item.show_stereotypes = button.get_active()


def stereotype_model(subject):
    model = Gtk.TreeStore.new(
        [
            str,  # stereotype/attribute
            str,  # value
            bool,  # is applied stereotype
            bool,  # show checkbox (is stereotype)
            bool,  # value editable
            object,  # stereotype / attribute
            object,  # value editable
            object,  # slot element
        ]
    )
    refresh(subject, model)

    return model, (_toggle_stereotype, subject, model), (_set_value, model)


def refresh(subject, model):
    stereotypes = UML.model.get_stereotypes(subject)
    instances = subject.appliedStereotype

    def upsert(path, parent, row_data):
        try:
            new_row = model.get_iter(path)
        except ValueError:
            new_row = model.append(parent, row_data)
        else:
            row = model[path]
            row[:] = row_data
        return new_row

    # shortcut map stereotype -> slot (InstanceSpecification)
    slots = {}
    for applied in instances:
        for slot in applied.slot:
            slots[slot.definingFeature] = slot

    for st_index, st in enumerate(stereotypes):
        for applied in instances:
            if st in applied.classifier:
                break
        else:
            applied = None

        parent = upsert(
            f"{st_index}",
            None,
            (st.name, "", bool(applied), True, False, st, None, None),
        )
        for attr_index, attr in enumerate(
            attr for attr in st.ownedAttribute if not attr.association
        ):
            slot = slots.get(attr)
            value = slot.value if slot else ""
            upsert(
                f"{st_index}:{attr_index}",
                parent,
                (
                    attr.name,
                    value,
                    bool(applied),
                    False,
                    bool(applied),
                    attr,
                    applied,
                    slot,
                ),
            )


@transactional
def _toggle_stereotype(renderer, path, subject, model):
    row = model[path]
    name, old_value, is_applied, _, _, stereotype, _, _ = row
    value = not is_applied

    if value:
        UML.model.apply_stereotype(subject, stereotype)
    else:
        UML.model.remove_stereotype(subject, stereotype)

    row[2] = value

    refresh(subject, model)


@transactional
def _set_value(renderer, path, value, model):
    """
    Set value of stereotype property applied to an UML element.

    Slot is created if instance Create valuChange value of instance spe
    """
    row = model[path]
    name, old_value, is_applied, _, _, attr, applied, slot = row
    if isinstance(attr, UML.Stereotype):
        return  # don't edit stereotype rows

    if slot is None and not value:
        return  # nothing to do and don't create slot without value

    if slot is None:
        slot = UML.model.add_slot(applied, attr)

    assert slot

    if value:
        slot.value = value
    else:
        # no value, then remove slot
        del applied.slot[slot]
        slot = None
        value = ""

    row[1] = value
    row[5] = slot



from gaphor.UML.components import (
    componentsgrouping,
    componentspropertypage,
    connectorconnect,
    copypaste,
)
from gaphor.UML.components.artifact import ArtifactItem
from gaphor.UML.components.component import ComponentItem
from gaphor.UML.components.connector import ConnectorItem
from gaphor.UML.components.node import NodeItem

__all__ = ["ArtifactItem", "ComponentItem", "ConnectorItem", "NodeItem"]



import gaphas
import pytest

from gaphor.core.eventmanager import EventManager
from gaphor.core.modeling import ElementFactory, Presentation
from gaphor.core.modeling.diagram import Diagram, StyledDiagram, StyledItem


@pytest.fixture
def event_manager():
    return EventManager()


@pytest.fixture
def element_factory(event_manager):
    return ElementFactory(event_manager)


@pytest.fixture
def diagram(element_factory):
    diagram = element_factory.create(Diagram)
    yield diagram
    diagram.unlink()


class DemoItem(Presentation, gaphas.Item):
    pass


def test_name_does_not_have_item_suffix(diagram):
    item = diagram.create(DemoItem)
    node = StyledItem(item)

    assert node.name() == "demo"


def test_diagram_is_parent_of_item(diagram):
    item = diagram.create(DemoItem)
    node = StyledItem(item)

    assert node.parent().name() == "diagram"


def test_diagram_has_no_parent(diagram):
    node = StyledDiagram(diagram)

    assert node.parent() is None



import pytest

from gaphor.core.modeling import ElementFactory
from gaphor.core.modeling.elementdispatcher import ElementDispatcher
from gaphor.diagram.tests.fixtures import diagram, event_manager  # noqa F401
from gaphor.SysML.modelinglanguage import SysMLModelingLanguage
from gaphor.UML.modelinglanguage import UMLModelingLanguage


class MockModelingLanguage:
    def __init__(self):
        self._modeling_languages = [UMLModelingLanguage(), SysMLModelingLanguage()]

    def lookup_element(self, name):
        return self.first(lambda provider: provider.lookup_element(name))

    def first(self, predicate):
        for provider in self._modeling_languages:
            type = predicate(provider)
            if type:
                return type


@pytest.fixture
def element_factory(event_manager):  # noqa: F811
    return ElementFactory(
        event_manager, ElementDispatcher(event_manager, MockModelingLanguage())
    )



"""Unit tests for simple items."""

from gaphas import View

from gaphor.diagram.general.simpleitem import Ellipse, Line
from gaphor.tests import TestCase


class SimpleItemTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.view = View(self.diagram.canvas)

    def test_line(self):
        """
        """
        self.diagram.create(Line)

    def test_box(self):
        """
        """
        self.diagram.create(Line)

    def test_ellipse(self):
        """
        """
        self.diagram.create(Ellipse)



"""This module contains some helpers that can be used to execute generator
functions in the GLib main loop.

This module provided the following classes:
GIdleThread - Thread like behavior for generators in a main loop
Queue - A simple queue implementation suitable for use with GIdleThread

Exceptions:
QueueEmpty - raised when one tried to get a value of an empty queue
QueueFull - raised when the queue reaches it's max size and the oldest item
            may not be disposed.
"""

import time
import types
from typing import List, Optional

from gi.repository import GLib


class GIdleThread:
    """This is a pseudo-"thread" for use with the GTK+ main loop.

    This class does act a bit like a thread, all code is executed in
    the callers thread though. The provided function should be a generator
    (or iterator).

    It can be started with start(). While the "thread" is running is_alive()
    can be called to see if it's alive. wait([timeout]) will wait till the
    generator is finished, or timeout seconds.

    If an exception is raised from within the generator, it is stored in
    the exc_info property. Execution of the generator is finished. The
    exc_info property contains a tuple (exc_type, exc_value, exc_traceback),
    see sys.exc_info() for details.

    Note that this routine runs in the current thread, so there is no need
    for nasty locking schemes.

    Example (runs a counter through the GLib main loop routine):
    >>> def counter(max):
    ...     for x in range(max):
    ...         yield x
    >>> t = GIdleThread(counter(123))
    >>> id = t.start()
    >>> main = GLib.main_context_default()
    >>> t.is_alive()
    True
    >>> main.iteration(False) # doctest: +ELLIPSIS
    True

    """

    def __init__(self, generator, queue=None):
        assert isinstance(
            generator, types.GeneratorType
        ), "The generator should be an iterator"
        self._generator = generator
        self._queue = queue
        self._idle_id = 0
        self._exc: Optional[BaseException] = None

    def start(self, priority=GLib.PRIORITY_LOW):
        """Start the generator. Default priority is low, so screen updates
        will be allowed to happen.
        """
        idle_id = GLib.idle_add(self.__generator_executer, priority=priority)
        self._idle_id = idle_id
        return idle_id

    def wait(self, timeout=0):
        """Waits until finished or timeout.

        Waits until the coroutine is finished or return after timeout seconds.
        This is achieved by running the GTK+ main loop.

        Args:
            timeout: A float timeout in seconds, default is no timeout.

        Returns:
            A bool, True for timeout, False for coroutine ending.
        """

        start_time = time.monotonic()
        main_ctx = GLib.main_context_default()
        while self.is_alive():
            main_ctx.iteration(False)
            if timeout and (time.monotonic() - start_time >= timeout):
                return True
        return False

    def interrupt(self):
        """Force the generator to stop running.
        """
        if self.is_alive():
            GLib.source_remove(self._idle_id)
            self._idle_id = 0

    def is_alive(self):
        """Returns True if the generator is still running.
        """
        return self._idle_id != 0

    error = property(
        lambda self: self._exc,
        doc="Return a possible exception that had occurred "
        "during execution of the generator",
    )

    def reraise(self):
        """Rethrow the error that occurred during execution of the idle process.
        """
        exc = self._exc

        if exc:
            raise exc

    def __generator_executer(self):
        try:
            result = next(self._generator)
            if self._queue:
                try:
                    self._queue.put(result)
                except QueueFull:
                    self.wait(0.5)
                    # If this doesn't work...
                    self._queue.put(result)
            return True
        except StopIteration:
            self._idle_id = 0
            return False
        except BaseException as e:
            self._exc = e
            self._idle_id = 0
            return False


class QueueEmpty(Exception):
    """Exception raised whenever the queue is empty and someone tries to fetch
    a value.
    """


class QueueFull(Exception):
    """Exception raised when the queue is full and the oldest item may not be
    disposed.
    """


class Queue:
    """A FIFO queue. If the queue has a max size, the oldest item on the
    queue is dropped if that size id exceeded.
    """

    def __init__(self, size=0, dispose_oldest=True):
        self._queue: List[object] = []
        self._size = size
        self._dispose_oldest = dispose_oldest

    def put(self, item):
        """Put item on the queue. If the queue size is limited ...
        """
        if self._size > 0 and len(self._queue) >= self._size:
            if self._dispose_oldest:
                self.get()
            else:
                raise QueueFull

        self._queue.insert(0, item)

    def get(self):
        """Get the oldest item off the queue.
        QueueEmpty is raised if no items are left on the queue.
        """
        try:
            return self._queue.pop()
        except IndexError:
            raise QueueEmpty


if __name__ == "__main__":

    def counter(max):
        yield from range(max)

    def shower(queue):
        # Never stop reading the queue:
        while True:
            try:
                cnt = queue.get()
                print("cnt =", cnt)
            except QueueEmpty:
                pass
            yield None

    print("Test 1: (should print range 0..22)")
    queue = Queue()
    c = GIdleThread(counter(23), queue)
    s = GIdleThread(shower(queue))

    main = GLib.main_context_default()
    c.start()
    s.start()
    s.wait(2)

    print("Test 2: (should only print 22)")
    queue = Queue(size=1)
    c = GIdleThread(counter(23), queue)
    s = GIdleThread(shower(queue))

    main = GLib.main_context_default()
    c.start(priority=GLib.PRIORITY_DEFAULT)
    s.start()
    s.wait(3)



import pytest

from gaphor.core.styling import parse_style_sheet
from gaphor.diagram.text import FontWeight


def first_decl_block(css):
    prop, value = next(parse_style_sheet(css))
    return value


@pytest.mark.parametrize(
    "color,rgba",
    [
        ["red", (1, 0, 0, 1)],
        ["#00ff00", (0, 1, 0, 1)],
        ["rgb(0, 255, 0)", (0, 1, 0, 1)],
        ["rgba(0, 255, 0, 0.3)", (0, 1, 0, 0.3)],
        ["rgb(0, 300, 0)", (0, 1, 0, 1)],
        ["gred", None],
        ["#00fg00", None],
        ["#00fg0000", None],
    ],
)
def test_color_parsing(color, rgba):
    css = f"* {{ color: {color}; }}"
    props = first_decl_block(css)

    if rgba is None:
        assert "color" not in props
    else:
        assert tuple(props["color"]) == rgba


def test_negative_number():
    css = "* { min-width: -1; }"
    props = first_decl_block(css)

    assert "min-width" not in props


def test_not_a_number():
    css = "* { min-width: foo; }"
    props = first_decl_block(css)

    assert "min-width" not in props


@pytest.mark.parametrize(
    "family,expected",
    [
        ["serif", "serif"],
        ["sans serif", "sans serif"],
        ['"DejaVu serif"', "DejaVu serif"],
        ["Number 1", "Number 1"],
        ["", None],
    ],
)
def test_string_style(family, expected):
    css = f"* {{ font-family: {family}; }}"
    props = first_decl_block(css)

    assert props.get("font-family") == expected


@pytest.mark.parametrize(
    "padding,expected",
    [
        [0, (0, 0, 0, 0)],
        [3, (3, 3, 3, 3)],
        ["1 3", (1, 3, 1, 3)],
        ["1 2 3", (1, 2, 3, 2)],
        ["1 2 3 4", (1, 2, 3, 4)],
        ["foo", None],
        ["", None],
    ],
)
def test_padding(padding, expected):
    css = f"* {{ padding: {padding}; }}"
    props = first_decl_block(css)

    assert props.get("padding") == expected


def test_enum_style():
    css = "* { font-weight: bold; }"
    props = first_decl_block(css)

    assert props["font-weight"] == FontWeight.BOLD


def test_multi_declaration():
    css = "* { font-weight: bold; color: black; }"
    props = first_decl_block(css)

    assert props["font-weight"] == FontWeight.BOLD
    assert props["color"] == (0, 0, 0, 1)



from gi.repository import Gtk

from gaphor import UML
from gaphor.core import transactional
from gaphor.diagram.propertypages import PropertyPageBase, PropertyPages, new_builder


def _issubclass(c, b):
    try:
        return issubclass(c, b)
    except TypeError:
        return False


@PropertyPages.register(UML.Class)
class MetaclassPropertyPage(PropertyPageBase):
    """Adapter which shows a property page for a class view.
    Also handles metaclasses.
    """

    order = 10

    subject: UML.Class

    CLASSES = list(
        sorted(
            c
            for c in dir(UML)
            if _issubclass(getattr(UML, c), UML.Element) and c != "Stereotype"
        )
    )

    def __init__(self, subject):
        self.subject = subject
        self.watcher = subject.watcher()

    def construct(self):
        if not UML.model.is_metaclass(self.subject):
            return

        builder = new_builder("metaclass-editor")

        combo = builder.get_object("metaclass-combo")
        for c in self.CLASSES:
            combo.append_text(c)

        completion = Gtk.EntryCompletion()
        completion.set_model(combo.get_model())
        completion.set_minimum_key_length(1)
        completion.set_text_column(0)

        entry = combo.get_child()
        entry.set_completion(completion)

        entry.set_text(self.subject and self.subject.name or "")

        def handler(event):
            if event.element is self.subject and event.new_value is not None:
                entry.set_text(event.new_value)

        self.watcher.watch("name", handler).subscribe_all()

        builder.connect_signals(
            {
                "metaclass-combo-changed": (self._on_name_changed,),
                "metaclass-combo-destroy": (self.watcher.unsubscribe_all,),
            }
        )

        return builder.get_object("metaclass-editor")

    @transactional
    def _on_name_changed(self, combo):
        self.subject.name = combo.get_active_text()



"""
Connector adapters.

To register connectors implemented in this module, it is imported in
gaphor.adapter package.
"""

from __future__ import annotations

from typing import List, Optional, Type, TypeVar, Union

from gaphas.canvas import Canvas, Connection
from gaphas.connector import Handle, Port
from generic.multidispatch import FunctionDispatcher, multidispatch
from typing_extensions import Protocol

from gaphor.core.modeling import Element, Presentation
from gaphor.core.modeling.properties import association, redefine, relation
from gaphor.diagram.presentation import ElementPresentation, LinePresentation

T = TypeVar("T", bound=Element)


class ConnectorProtocol(Protocol):
    def __init__(self, element: object, line: object,) -> None:
        ...

    def allow(self, handle: Handle, port: Port) -> bool:
        ...

    def connect(self, handle: Handle, port: Port) -> bool:
        ...

    def disconnect(self, handle: Handle) -> None:
        ...


class BaseConnector:
    """
    Connection adapter for Gaphor diagram items.

    Line item ``line`` connects with a handle to a connectable item ``element``.

    :param line: connecting item
    :type line: Presentation
    :param element: connectable item
    :type element: Presentation

    The following methods are required to make this work:

    - ``allow()``: is the connection allowed at all (during mouse movement for example).
    - ``connect()``: Establish a connection between element and line. Also takes care of
      disconnects, if required (e.g. 1:1 relationships)
    - ``disconnect()``: Break connection, called when dropping a handle on a
      point where it can not connect.
    - ``reconnect()`` (*Optional*): Connect to another item (only used if present)

    By convention the adapters are registered by (element, line) -- in that order.
    """

    def __init__(
        self, element: Presentation[Element], line: Presentation[Element],
    ) -> None:
        assert element.canvas is line.canvas
        self.element = element
        self.line = line
        self.canvas: Canvas = element.canvas

    def get_connection(self, handle: Handle) -> Optional[Connection]:
        """
        Get connection information
        """
        return self.canvas.get_connection(handle)

    def get_connected(self, handle: Handle) -> Optional[Presentation[Element]]:
        """
        Get item connected to a handle.
        """
        cinfo = self.canvas.get_connection(handle)
        if cinfo:
            return cinfo.connected  # type: ignore[no-any-return] # noqa: F723
        return None

    def allow(self, handle: Handle, port: Port) -> bool:
        """
        Determine if items can be connected.

        Returns `True` if connection is allowed.
        """
        return True

    def connect(self, handle: Handle, port: Port) -> bool:
        """
        Connect to an element. Note that at this point the line may
        be connected to some other, or the same element.
        Also the connection at model level still exists.

        Returns `True` if a connection is established.
        """
        return True

    def disconnect(self, handle: Handle) -> None:
        """Disconnect model level connections."""


class NoConnector:
    def __init__(self, element, line,) -> None:
        pass

    def allow(self, handle: Handle, port: Port) -> bool:
        return False

    def connect(self, handle: Handle, port: Port) -> bool:
        return False

    def disconnect(self, handle: Handle) -> None:
        pass


# Work around issue https://github.com/python/mypy/issues/3135 (Class decorators are not type checked)
# This definition, along with the the ignore below, seems to fix the behaviour for mypy at least.
Connector: FunctionDispatcher[Type[ConnectorProtocol]] = multidispatch(object, object)(
    NoConnector
)


class UnaryRelationshipConnect(BaseConnector):
    """
    Base class for relationship connections, such as associations,
    dependencies and implementations.

    Unary relationships are allowed to connect both ends to the same element

    This class introduces a new method: relationship() which is used to
    find an existing relationship in the model that does not yet exist
    on the canvas.
    """

    element: ElementPresentation[Element]
    line: LinePresentation[Element]

    def relationship(
        self, required_type: Type[Element], head: relation, tail: relation
    ) -> Optional[Element]:
        """
        Find an existing relationship in the model that meets the
        required type and is connected to the same model element the head
        and tail of the line are connected to.

        type - the type of relationship we're looking for
        head - tuple (association name on line, association name on element)
        tail - tuple (association name on line, association name on element)
        """
        line = self.line

        line_head = self.get_connected(line.head)
        line_tail = self.get_connected(line.tail)
        assert line_head
        assert line_tail
        head_subject = line_head.subject
        tail_subject = line_tail.subject

        # First check if the right subject is already connected:
        if (
            line.subject
            and getattr(line.subject, head.name) is head_subject
            and getattr(line.subject, tail.name) is tail_subject
        ):
            return line.subject

        assert isinstance(head, (association, redefine)), f"head is {head}"
        assert isinstance(tail, (association, redefine)), f"tail is {tail}"

        # Try to find a relationship, that is already created, but not
        # yet displayed in the diagram on the tail side, since tail should
        # have a reference to head at least.
        if not tail.opposite:
            return None

        gen: Element
        for gen in getattr(tail_subject, tail.opposite):
            if not isinstance(gen, required_type):
                continue

            gen_head = getattr(gen, head.name)
            try:
                if head_subject not in gen_head:
                    continue
            except TypeError:
                if gen_head is not head_subject:
                    continue

            # Check for this entry on line.canvas
            item: Union[ElementPresentation, LinePresentation]
            for item in gen.presentation:
                # Allow line to be returned. Avoids strange
                # behaviour during loading
                if item.canvas is line.canvas and item is not line:
                    break
            else:
                return gen
        return None

    def relationship_or_new(self, type: Type[T], head: relation, tail: relation) -> T:
        """
        Like relation(), but create a new instance if none was found.
        """
        relation = self.relationship(type, head, tail)
        if not relation:
            line = self.line
            relation = line.model.create(type)
            line_head = self.get_connected(line.head)
            line_tail = self.get_connected(line.tail)
            assert line_head
            assert line_tail
            setattr(relation, head.name, line_head.subject)
            setattr(relation, tail.name, line_tail.subject)
        assert isinstance(relation, type)
        return relation

    def reconnect_relationship(
        self, handle: Handle, head: relation, tail: relation
    ) -> None:
        """
        Reconnect relationship for given handle.

        :Parameters:
         handle
            Handle at which reconnection happens.
         head
            Relationship head attribute name.
         tail
            Relationship tail attribute name.
        """
        line = self.line
        c1 = self.get_connected(line.head)
        c2 = self.get_connected(line.tail)
        assert c1
        assert c2
        if line.head is handle:
            setattr(line.subject, head.name, c1.subject)
        elif line.tail is handle:
            setattr(line.subject, tail.name, c2.subject)
        else:
            raise ValueError("Incorrect handle passed to adapter")

    def connect_connected_items(self, connections: None = None) -> None:
        """
        Cause items connected to ``line`` to reconnect, allowing them to
        establish or destroy relationships at model level.
        """
        line = self.line
        canvas = self.canvas
        solver = canvas.solver

        # First make sure coordinates match
        solver.solve()
        for cinfo in connections or canvas.get_connections(connected=line):
            if line is cinfo.connected:
                continue
            adapter = Connector(line, cinfo.connected)
            assert adapter, "No element to connect {} and {}".format(
                line, cinfo.connected
            )
            adapter.connect(cinfo.handle, cinfo.port)

    def disconnect_connected_items(self) -> List[Connection]:
        """
        Cause items connected to @line to be disconnected.
        This is necessary if the subject of the @line is to be removed.

        Returns a list of (item, handle) pairs that were connected (this
        list can be used to connect items again with connect_connected_items()).
        """
        line = self.line
        canvas = self.canvas
        solver = canvas.solver

        # First make sure coordinates match
        solver.solve()
        connections = list(canvas.get_connections(connected=line))
        for cinfo in connections:
            adapter = Connector(cinfo.item, cinfo.connected)
            adapter.disconnect(cinfo.handle)
        return connections

    def connect_subject(self, handle: Handle) -> bool:
        """
        Establish the relationship at model level.
        """
        raise NotImplementedError("Implement connect_subject() in a subclass")

    def disconnect_subject(self, handle: Handle) -> None:
        """
        Disconnect the diagram item from its model element. If there are
        no more presentations(diagram items) connected to the model element,
        unlink() it too.
        """
        line = self.line
        old = line.subject
        del line.subject
        if old and len(old.presentation) == 0:
            old.unlink()

    def connect(self, handle: Handle, port: Port) -> bool:
        """
        Connect the items to each other. The model level relationship
        is created by create_subject()
        """
        if not super().connect(handle, port):
            return False
        opposite = self.line.opposite(handle)
        oct = self.get_connected(opposite)
        if oct:
            self.connect_subject(handle)
            line = self.line
            if line.subject:
                self.connect_connected_items()
        return True

    def disconnect(self, handle: Handle) -> None:
        """
        Disconnect model element.
        """
        line = self.line
        opposite = line.opposite(handle)
        oct = self.get_connected(opposite)
        hct = self.get_connected(handle)

        if hct and oct:
            # Both sides of line are connected => disconnect
            self.disconnect_connected_items()
            self.disconnect_subject(handle)

        super().disconnect(handle)


class RelationshipConnect(UnaryRelationshipConnect):
    """Base for relationship connections between unique elements."""

    def allow(self, handle: Handle, port: Port) -> bool:
        """
        In addition to the normal check, both relationship ends may not be
        connected to the same element. Same goes for subjects.
        """
        opposite = self.line.opposite(handle)
        element = self.element
        connected_to = self.get_connected(opposite)

        # Element can not be a parent of itself.
        if connected_to is element:
            return False

        # Same goes for subjects:
        if (
            connected_to
            and not connected_to.subject
            and not element.subject
            and connected_to.subject is element.subject
        ):
            return False

        return super().allow(handle, port)



"""
Undo management for Gaphor.

Undoing and redoing actions is managed through the UndoManager.

An undo action should be a callable object (called with no arguments).

An undo action should return a callable object that acts as redo function.
If None is returned the undo action is considered to be the redo action as well.

NOTE: it would be nice to use actions in conjunction with functools.partial.
"""

import logging
from typing import Callable, List

from gaphas import state

from gaphor.abc import ActionProvider, Service
from gaphor.action import action
from gaphor.core import event_handler
from gaphor.core.modeling.event import (
    AssociationAdded,
    AssociationDeleted,
    AssociationSet,
    AttributeUpdated,
    ElementCreated,
    ElementDeleted,
    ModelReady,
)
from gaphor.core.modeling.properties import association as association_property
from gaphor.event import (
    ActionEnabled,
    ServiceEvent,
    TransactionBegin,
    TransactionCommit,
    TransactionRollback,
)
from gaphor.transaction import Transaction, transactional

logger = logging.getLogger(__name__)


class ActionStack:
    """
    A transaction. Every action that is added between a begin_transaction()
    and a commit_transaction() call is recorded in a transaction, so it can
    be played back when a transaction is executed. This executing a
    transaction has the effect of performing the actions recorded, which will
    typically undo actions performed by the user.
    """

    def __init__(self):
        self._actions: List[Callable[[], None]] = []

    def add(self, action):
        self._actions.append(action)

    def can_execute(self):
        return self._actions and True or False

    @transactional
    def execute(self):
        self._actions.reverse()

        for act in self._actions:
            try:
                act()
            except Exception:
                logger.error(f"Error while undoing action {act}", exc_info=True)


class UndoManagerStateChanged(ServiceEvent):
    """
    Event class used to send state changes on the Undo Manager.
    """

    def __init__(self, service):
        self.service = service


class UndoManager(Service, ActionProvider):
    """
    Simple transaction manager for Gaphor.
    This transaction manager supports nested transactions.

    The Undo manager sports an undo and a redo stack. Each stack contains
    a set of actions that can be executed, just by calling them (e.i action())
    If something is returned by an action, that is considered the callable
    to be used to undo or redo the last performed action.
    """

    def __init__(self, event_manager):
        self.event_manager = event_manager
        self._undo_stack: List[ActionStack] = []
        self._redo_stack: List[ActionStack] = []
        self._stack_depth = 20
        self._current_transaction = None

        event_manager.subscribe(self.reset)
        event_manager.subscribe(self.begin_transaction)
        event_manager.subscribe(self.commit_transaction)
        event_manager.subscribe(self.rollback_transaction)
        self._register_undo_handlers()
        self._action_executed()

    def shutdown(self):
        self.event_manager.unsubscribe(self.reset)
        self.event_manager.unsubscribe(self.begin_transaction)
        self.event_manager.unsubscribe(self.commit_transaction)
        self.event_manager.unsubscribe(self.rollback_transaction)
        self._unregister_undo_handlers()

    def clear_undo_stack(self):
        self._undo_stack = []
        self._current_transaction = None

    def clear_redo_stack(self):
        del self._redo_stack[:]

    @event_handler(ModelReady)
    def reset(self, event=None):
        self.clear_redo_stack()
        self.clear_undo_stack()
        self._action_executed()

    @event_handler(TransactionBegin)
    def begin_transaction(self, event=None):
        """
        Add an action to the current transaction
        """
        assert not self._current_transaction
        self._current_transaction = ActionStack()

    def add_undo_action(self, action):
        """
        Add an action to undo.
        """
        if self._current_transaction:
            self._current_transaction.add(action)
            # TODO: should this be placed here?
            self._action_executed()

    @event_handler(TransactionCommit)
    def commit_transaction(self, event=None):
        assert self._current_transaction

        if self._current_transaction.can_execute():
            # Here:
            self.clear_redo_stack()
            self._undo_stack.append(self._current_transaction)
            while len(self._undo_stack) > self._stack_depth:
                del self._undo_stack[0]

        self._current_transaction = None

        self._action_executed()

    @event_handler(TransactionRollback)
    def rollback_transaction(self, event=None):
        """
        Roll back the transaction we're in.
        """
        assert self._current_transaction

        # Store stacks
        undo_stack = list(self._undo_stack)

        erroneous_tx = self._current_transaction
        self._current_transaction = None
        try:
            with Transaction(self.event_manager):
                try:
                    erroneous_tx.execute()
                except Exception as e:
                    logger.error("Could not roolback transaction")
                    logger.error(e)
        finally:
            # Discard all data collected in the rollback "transaction"
            self._undo_stack = undo_stack

        self._action_executed()

    def discard_transaction(self):

        self._current_transaction = None

        self._action_executed()

    @action(name="edit-undo", shortcut="<Primary>z")
    def undo_transaction(self):
        if not self._undo_stack:
            return

        if self._current_transaction:
            logger.warning("Trying to undo a transaction, while in a transaction")
            self.commit_transaction()
        transaction = self._undo_stack.pop()

        # Store stacks
        undo_stack = list(self._undo_stack)
        redo_stack = list(self._redo_stack)
        self._undo_stack = []

        try:
            with Transaction(self.event_manager):
                transaction.execute()
        finally:
            # Restore stacks and put latest tx on the redo stack
            self._redo_stack = redo_stack
            if self._undo_stack:
                self._redo_stack.extend(self._undo_stack)
            self._undo_stack = undo_stack

        while len(self._redo_stack) > self._stack_depth:
            del self._redo_stack[0]

        self._action_executed()

    @action(
        name="edit-redo", shortcut="<Primary><Shift>z",
    )
    def redo_transaction(self):
        if not self._redo_stack:
            return

        transaction = self._redo_stack.pop()

        redo_stack = list(self._redo_stack)
        try:
            with Transaction(self.event_manager):
                transaction.execute()
        finally:
            self._redo_stack = redo_stack

        self._action_executed()

    def in_transaction(self):
        return self._current_transaction is not None

    def can_undo(self):
        return bool(self._current_transaction or self._undo_stack)

    def can_redo(self):
        return bool(self._redo_stack)

    def _action_executed(self, event=None):
        self.event_manager.handle(ActionEnabled("win.edit-undo", self.can_undo()))
        self.event_manager.handle(ActionEnabled("win.edit-redo", self.can_redo()))
        self.event_manager.handle(UndoManagerStateChanged(self))

    #
    # Undo Handlers
    #

    def _gaphas_undo_handler(self, event):
        self.add_undo_action(lambda: state.saveapply(*event))

    def _register_undo_handlers(self):

        logger.debug("Registering undo handlers")

        self.event_manager.subscribe(self.undo_create_event)
        self.event_manager.subscribe(self.undo_delete_event)
        self.event_manager.subscribe(self.undo_attribute_change_event)
        self.event_manager.subscribe(self.undo_association_set_event)
        self.event_manager.subscribe(self.undo_association_add_event)
        self.event_manager.subscribe(self.undo_association_delete_event)

        #
        # Direct revert-statements from gaphas to the undomanager
        state.observers.add(state.revert_handler)

        state.subscribers.add(self._gaphas_undo_handler)

    def _unregister_undo_handlers(self):

        logger.debug("Unregistering undo handlers")

        self.event_manager.unsubscribe(self.undo_create_event)
        self.event_manager.unsubscribe(self.undo_delete_event)
        self.event_manager.unsubscribe(self.undo_attribute_change_event)
        self.event_manager.unsubscribe(self.undo_association_set_event)
        self.event_manager.unsubscribe(self.undo_association_add_event)
        self.event_manager.unsubscribe(self.undo_association_delete_event)

        state.observers.discard(state.revert_handler)

        state.subscribers.discard(self._gaphas_undo_handler)

    @event_handler(ElementCreated)
    def undo_create_event(self, event):
        factory = event.service
        # A factory is not always present, e.g. for DiagramItems
        if not factory:
            return
        element = event.element

        def _undo_create_event():
            try:
                del factory._elements[element.id]
            except KeyError:
                pass  # Key was probably already removed in an unlink call
            self.event_manager.handle(ElementDeleted(factory, element))

        self.add_undo_action(_undo_create_event)

    @event_handler(ElementDeleted)
    def undo_delete_event(self, event):
        factory = event.service
        # A factory is not always present, e.g. for DiagramItems
        if not factory:
            return
        element = event.element
        assert factory, f"No factory defined for {element} ({factory})"

        def _undo_delete_event():
            factory._elements[element.id] = element
            self.event_manager.handle(ElementCreated(factory, element))

        self.add_undo_action(_undo_delete_event)

    @event_handler(AttributeUpdated)
    def undo_attribute_change_event(self, event):
        attribute = event.property
        element = event.element
        value = event.old_value

        def _undo_attribute_change_event():
            attribute._set(element, value)

        self.add_undo_action(_undo_attribute_change_event)

    @event_handler(AssociationSet)
    def undo_association_set_event(self, event):
        association = event.property
        if type(association) is not association_property:
            return
        element = event.element
        value = event.old_value

        def _undo_association_set_event():
            # Tell the association it should not need to let the opposite
            # side connect (it has it's own signal)
            association._set(element, value, from_opposite=True)

        self.add_undo_action(_undo_association_set_event)

    @event_handler(AssociationAdded)
    def undo_association_add_event(self, event):
        association = event.property
        if type(association) is not association_property:
            return
        element = event.element
        value = event.new_value

        def _undo_association_add_event():
            # Tell the association it should not need to let the opposite
            # side connect (it has it's own signal)
            association._del(element, value, from_opposite=True)

        self.add_undo_action(_undo_association_add_event)

    @event_handler(AssociationDeleted)
    def undo_association_delete_event(self, event):
        association = event.property
        if type(association) is not association_property:
            return
        element = event.element
        value = event.old_value

        def _undo_association_delete_event():
            # Tell the assoctaion it should not need to let the opposite
            # side connect (it has it's own signal)
            association._set(element, value, from_opposite=True)

        self.add_undo_action(_undo_association_delete_event)



"""Test Metaclass Property Page."""

import pytest
from gi.repository import Gtk

from gaphor import UML
from gaphor.UML.profiles.metaclasspropertypage import MetaclassPropertyPage


@pytest.fixture
def class_(element_factory):
    class_ = element_factory.create(UML.Class)
    class_.name = "Class"
    yield class_
    del class_


def test_name_input_field_for_normal_class(class_):
    """Test Metaclass Property Page not create for normal Class."""

    editor = MetaclassPropertyPage(class_)
    page = editor.construct()

    assert not page


def test_name_selection_for_metaclass(element_factory, class_):
    """Test the creation of Metaclass Property Page."""

    stereotype = element_factory.create(UML.Stereotype)
    stereotype.name = "NewStereotype"
    UML.model.create_extension(class_, stereotype)

    editor = MetaclassPropertyPage(class_)
    page = editor.construct()

    assert page

    combo = page.get_children()[1]

    assert type(combo) is Gtk.ComboBoxText
    assert combo.get_child().get_text() == "Class"

    class_.name = "Blah"

    assert combo.get_child().get_text() == "Blah"



"""
Test transition item and state vertices connections.
"""

from gaphor import UML
from gaphor.tests import TestCase
from gaphor.UML.states.finalstate import FinalStateItem
from gaphor.UML.states.pseudostates import PseudostateItem
from gaphor.UML.states.state import StateItem
from gaphor.UML.states.transition import TransitionItem


class TransitionConnectorTestCase(TestCase):

    services = TestCase.services

    def test_vertex_connect(self):
        """Test transition to state vertex connection
        """
        v1 = self.create(StateItem, UML.State)
        v2 = self.create(StateItem, UML.State)

        t = self.create(TransitionItem)
        assert t.subject is None

        # connect vertices with transition
        self.connect(t, t.head, v1)
        self.connect(t, t.tail, v2)

        assert t.subject is not None

        assert len(self.kindof(UML.Transition)) == 1

        assert t.subject == v1.subject.outgoing[0]
        assert t.subject == v2.subject.incoming[0]
        assert t.subject.source == v1.subject
        assert t.subject.target == v2.subject

    def test_vertex_reconnect(self):
        """Test transition to state vertex reconnection
        """
        v1 = self.create(StateItem, UML.State)
        v2 = self.create(StateItem, UML.State)
        v3 = self.create(StateItem, UML.State)

        t = self.create(TransitionItem)

        # connect: v1 -> v2
        self.connect(t, t.head, v1)
        self.connect(t, t.tail, v2)

        s = t.subject
        s.name = "tname"
        s.guard.specification = "tguard"

        # reconnect: v1 -> v3
        self.connect(t, t.tail, v3)

        assert s is t.subject
        assert len(self.kindof(UML.Transition)) == 1

        assert t.subject == v1.subject.outgoing[0]
        assert t.subject == v3.subject.incoming[0]
        assert t.subject.source == v1.subject
        assert t.subject.target == v3.subject

        assert len(v2.subject.incoming) == 0
        assert len(v2.subject.outgoing) == 0

    def test_vertex_disconnect(self):
        """Test transition and state vertices disconnection
        """
        t = self.create(TransitionItem)
        v1 = self.create(StateItem, UML.State)
        v2 = self.create(StateItem, UML.State)

        self.connect(t, t.head, v1)
        self.connect(t, t.tail, v2)
        assert t.subject is not None

        assert len(self.kindof(UML.Transition)) == 1

        # test preconditions
        assert t.subject == v1.subject.outgoing[0]
        assert t.subject == v2.subject.incoming[0]

        self.disconnect(t, t.tail)
        assert t.subject is None

        self.disconnect(t, t.head)
        assert t.subject is None

    def test_initial_pseudostate_connect(self):
        """Test transition and initial pseudostate connection
        """
        v1 = self.create(PseudostateItem, UML.Pseudostate)
        v2 = self.create(StateItem, UML.State)

        t = self.create(TransitionItem)
        assert t.subject is None

        # connect head of transition to an initial pseudostate
        self.connect(t, t.head, v1)
        assert t.subject is None

        # connect tail of transition to a state
        self.connect(t, t.tail, v2)
        assert t.subject is not None

        assert len(self.kindof(UML.Transition)) == 1

        # test preconditions
        assert t.subject == v1.subject.outgoing[0]
        assert t.subject == v2.subject.incoming[0]

        # we should not be able to connect two transitions to initial
        # pseudostate
        t2 = self.create(TransitionItem)
        # connection to `t2` should not be possible as v1 is already connected
        # to `t`
        glued = self.allow(t2, t2.head, v1)
        assert not glued
        assert self.get_connected(t2.head) is None

    def test_initial_pseudostate_disconnect(self):
        """Test transition and initial pseudostate disconnection
        """
        v1 = self.create(PseudostateItem, UML.Pseudostate)
        self.create(StateItem, UML.State)

        t = self.create(TransitionItem)
        assert t.subject is None

        # connect head of transition to an initial pseudostate
        self.connect(t, t.head, v1)
        assert self.get_connected(t.head)

        # perform the test
        self.disconnect(t, t.head)
        assert not self.get_connected(t.head)

    def test_initial_pseudostate_tail_glue(self):
        """Test transition tail and initial pseudostate gluing."""

        v1 = self.create(PseudostateItem, UML.Pseudostate)
        t = self.create(TransitionItem)
        assert t.subject is None

        # no tail connection should be possible
        glued = self.allow(t, t.tail, v1)
        assert not glued

    def test_final_state_connect(self):
        """Test transition to final state connection
        """
        v1 = self.create(StateItem, UML.State)
        v2 = self.create(FinalStateItem, UML.FinalState)
        t = self.create(TransitionItem)

        # connect head of transition to a state
        self.connect(t, t.head, v1)

        # check if transition can connect to final state
        glued = self.allow(t, t.tail, v2)
        assert glued
        # and connect tail of transition to final state
        self.connect(t, t.tail, v2)
        assert t.subject is not None

        assert len(self.kindof(UML.Transition)) == 1

        assert t.subject == v1.subject.outgoing[0]
        assert t.subject == v2.subject.incoming[0]
        assert t.subject.source == v1.subject
        assert t.subject.target == v2.subject

    def test_final_state_head_glue(self):
        """Test transition head to final state connection
        """
        v = self.create(FinalStateItem, UML.FinalState)
        t = self.create(TransitionItem)

        glued = self.allow(t, t.head, v)
        assert not glued



"""
Base code for presentation elements
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Generic, List, Optional, TypeVar

from gaphor.core.modeling import Element
from gaphor.core.modeling.properties import association, relation_one

if TYPE_CHECKING:
    from gaphas.canvas import Canvas  # noqa
    from gaphas.connector import Handle  # noqa
    from gaphas.matrix import Matrix  # noqa

S = TypeVar("S", bound=Element)


class Presentation(Element, Generic[S]):
    """
    This presentation is used to link the behaviors of `gaphor.core.modeling` and `gaphas.Item`.
    """

    def __init__(self, id=None, model=None):
        super().__init__(id, model)

        def update(event):
            self.request_update()

        self._watcher = self.watcher(default_handler=update)

        self.watch("subject")

    subject: relation_one[S] = association(
        "subject", Element, upper=1, opposite="presentation"
    )

    handles: Callable[[Presentation], List[Handle]]
    request_update: Callable[[Presentation], None]

    canvas: Optional[Canvas]

    matrix: Matrix

    @property
    def diagram(self):
        canvas = self.canvas
        return canvas.diagram if canvas else None

    def watch(self, path, handler=None):
        """
        Watch a certain path of elements starting with the DiagramItem.
        The handler is optional and will default to a simple
        self.request_update().

        Watches should be set in the constructor, so they can be registered
        and unregistered in one shot.

        This interface is fluent(returns self).
        """
        self._watcher.watch(path, handler)
        return self

    def subscribe_all(self):
        """
        Subscribe all watched paths, as defined through `watch()`.
        """
        self._watcher.subscribe_all()

    def unsubscribe_all(self):
        """
        Unsubscribe all watched paths, as defined through `watch()`.
        """
        self._watcher.unsubscribe_all()

    def unlink(self):
        """
        Remove the item from the canvas and set subject to None.
        """
        if self.canvas:
            self.canvas.remove(self)
        super().unlink()


Element.presentation = association(
    "presentation", Presentation, composite=True, opposite="subject"
)



"""
Test implementation (interface realization) item connectors.
"""

from gaphor import UML
from gaphor.tests import TestCase
from gaphor.UML.classes.implementation import ImplementationItem
from gaphor.UML.classes.interface import InterfaceItem
from gaphor.UML.classes.klass import ClassItem


class ImplementationTestCase(TestCase):
    def test_non_interface_glue(self):
        """Test non-interface gluing with implementation."""

        impl = self.create(ImplementationItem)
        clazz = self.create(ClassItem, UML.Class)

        glued = self.allow(impl, impl.head, clazz)
        # connecting head to non-interface item is disallowed
        self.assertFalse(glued)

    def test_interface_glue(self):
        """Test interface gluing with implementation
        """
        iface = self.create(InterfaceItem, UML.Interface)
        impl = self.create(ImplementationItem)

        glued = self.allow(impl, impl.head, iface)
        assert glued

    def test_classifier_glue(self):
        """Test classifier gluing with implementation
        """
        impl = self.create(ImplementationItem)
        clazz = self.create(ClassItem, UML.Class)

        glued = self.allow(impl, impl.tail, clazz)
        assert glued

    def test_connection(self):
        """Test connection of class and interface with implementation
        """
        iface = self.create(InterfaceItem, UML.Interface)
        impl = self.create(ImplementationItem)
        clazz = self.create(ClassItem, UML.Class)

        self.connect(impl, impl.head, iface)
        self.connect(impl, impl.tail, clazz)

        # check the datamodel
        self.assertTrue(isinstance(impl.subject, UML.Implementation))
        ct = self.get_connected(impl.head)
        assert ct is iface
        assert impl.subject is not None
        assert impl.subject.contract[0] is iface.subject
        assert impl.subject.implementatingClassifier[0] is clazz.subject

    def test_reconnection(self):
        """Test reconnection of class and interface with implementation
        """
        iface = self.create(InterfaceItem, UML.Interface)
        c1 = self.create(ClassItem, UML.Class)
        c2 = self.create(ClassItem, UML.Class)
        impl = self.create(ImplementationItem)

        # connect: iface -> c1
        self.connect(impl, impl.head, iface)
        self.connect(impl, impl.tail, c1)

        s = impl.subject

        # reconnect: iface -> c2
        self.connect(impl, impl.tail, c2)

        assert s is impl.subject
        assert len(impl.subject.contract) == 1
        assert len(impl.subject.implementatingClassifier) == 1
        assert iface.subject in impl.subject.contract
        assert c2.subject in impl.subject.implementatingClassifier
        assert (
            c1.subject not in impl.subject.implementatingClassifier
        ), impl.subject.implementatingClassifier
